# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for Queue resources."""
import uuid
from typing import cast
from urllib.parse import unquote

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import inject
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.restapi.routes import V1_QUEUES_ROUTE
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.schemas import IdStatusResponseSchema
from dioptra.restapi.v1.shared.drafts.controller import (
    generate_resource_drafts_endpoint,
    generate_resource_drafts_id_endpoint,
    generate_resource_id_draft_endpoint,
)
from dioptra.restapi.v1.shared.snapshots.controller import (
    generate_resource_snapshots_endpoint,
    generate_resource_snapshots_id_endpoint,
)
from dioptra.restapi.v1.shared.tags.controller import (
    generate_resource_tags_endpoint,
    generate_resource_tags_id_endpoint,
)

from .schema import (
    QueueGetQueryParameters,
    QueueMutableFieldsSchema,
    QueuePageSchema,
    QueueSchema,
)
from .service import RESOURCE_TYPE, SEARCHABLE_FIELDS, QueueIdService, QueueService

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace("Queues", description="Queues endpoint")


@api.route("/")
class QueueEndpoint(Resource):
    @inject
    def __init__(self, queue_service: QueueService, *args, **kwargs) -> None:
        """Initialize the queue resource.

        All arguments are provided via dependency injection.

        Args:
            queue_service: A QueueService object.
        """
        self._queue_service = queue_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=QueueGetQueryParameters, api=api)
    @responds(schema=QueuePageSchema, api=api)
    def get(self):
        """Gets a list of all Queue resources."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Queue", request_type="GET"
        )
        parsed_query_params = request.parsed_query_params  # noqa: F841

        group_id = parsed_query_params["group_id"]
        search_string = unquote(parsed_query_params["search"])
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]
        sort_by_string = parsed_query_params["sort_by"]
        descending = parsed_query_params["descending"]

        queues, total_num_queues = self._queue_service.get(
            group_id=group_id,
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
            sort_by_string=sort_by_string,
            descending=descending,
            log=log,
        )
        return utils.build_paging_envelope(
            V1_QUEUES_ROUTE,
            build_fn=utils.build_queue,
            data=queues,
            group_id=group_id,
            query=search_string,
            draft_type=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_queues,
            sort_by=sort_by_string,
            descending=descending,
        )

    @login_required
    @accepts(schema=QueueSchema, api=api)
    @responds(schema=QueueSchema, api=api)
    def post(self):
        """Creates a Queue resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Queue", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # noqa: F841

        queue = self._queue_service.create(
            name=parsed_obj["name"],
            description=parsed_obj["description"],
            group_id=parsed_obj["group_id"],
            log=log,
        )
        return utils.build_queue(queue)


@api.route("/<int:id>")
@api.param("id", "ID for the Queue resource.")
class QueueIdEndpoint(Resource):
    @inject
    def __init__(self, queue_id_service: QueueIdService, *args, **kwargs) -> None:
        """Initialize the queue resource.

        All arguments are provided via dependency injection.

        Args:
            queue_id_service: A QueueIdService object.
        """
        self._queue_id_service = queue_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=QueueSchema, api=api)
    def get(self, id: int):
        """Gets a Queue resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Queue", request_type="GET", id=id
        )
        queue = cast(
            models.Queue,
            self._queue_id_service.get(id, error_if_not_found=True, log=log),
        )
        return utils.build_queue(queue)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int):
        """Deletes a Queue resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Queue", request_type="DELETE", id=id
        )
        return self._queue_id_service.delete(queue_id=id, log=log)

    @login_required
    @accepts(schema=QueueMutableFieldsSchema, api=api)
    @responds(schema=QueueSchema, api=api)
    def put(self, id: int):
        """Modifies a Queue resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Queue", request_type="PUT", id=id
        )
        parsed_obj = request.parsed_obj  # type: ignore
        queue = cast(
            models.Queue,
            self._queue_id_service.modify(
                id,
                name=parsed_obj["name"],
                description=parsed_obj["description"],
                error_if_not_found=True,
                log=log,
            ),
        )
        return utils.build_queue(queue)


QueueDraftResource = generate_resource_drafts_endpoint(
    api,
    resource_name=RESOURCE_TYPE,
    route_prefix=V1_QUEUES_ROUTE,
    request_schema=QueueSchema,
)
QueueDraftIdResource = generate_resource_drafts_id_endpoint(
    api,
    resource_name=RESOURCE_TYPE,
    request_schema=QueueMutableFieldsSchema,
)
QueueIdDraftResource = generate_resource_id_draft_endpoint(
    api,
    resource_name=RESOURCE_TYPE,
    request_schema=QueueMutableFieldsSchema,
)

QueueSnapshotsResource = generate_resource_snapshots_endpoint(
    api=api,
    resource_model=models.Queue,
    resource_name=RESOURCE_TYPE,
    route_prefix=V1_QUEUES_ROUTE,
    searchable_fields=SEARCHABLE_FIELDS,
    page_schema=QueuePageSchema,
    build_fn=utils.build_queue,
)
QueueSnapshotsIdResource = generate_resource_snapshots_id_endpoint(
    api=api,
    resource_model=models.Queue,
    resource_name=RESOURCE_TYPE,
    response_schema=QueueSchema,
    build_fn=utils.build_queue,
)

QueueTagsResource = generate_resource_tags_endpoint(
    api=api,
    resource_name=RESOURCE_TYPE,
)
QueueTagsIdResource = generate_resource_tags_id_endpoint(
    api=api,
    resource_name=RESOURCE_TYPE,
)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
#
# This is a fork of the work
# https://github.com/gahjelle/pyplugs/blob/6921de46a2158462dc07c2f013155b53fbcebebb/tests/test_plugins.py  # noqa: B950
# See copyright below.
#
# Copyright (c) 2019 Geir Arne Hjelle
# Distributed under the terms of the MIT License.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Tests for Pyplugs

Based on the Pytest test runner
"""
import importlib
import pathlib
import sys

import pytest
from prefect import Flow

from dioptra import pyplugs
from dioptra.sdk.exceptions import (
    UnknownPackageError,
    UnknownPluginError,
    UnknownPluginFunctionError,
)


@pytest.fixture
def tmpfile(tmpdir):
    """A temporary file that can be read"""
    file_path = tmpdir.join("test")
    file_path.write("Temporary test file")

    return file_path


@pytest.fixture
def plugin_package():
    """Name of the test plugin package"""
    plugins_dir = pathlib.Path(__file__).absolute().parent / "plugin_directory"
    relative = plugins_dir.relative_to(pathlib.Path.cwd())

    return ".".join(relative.parts)


@pytest.fixture
def pyplugs_no_prefect(monkeypatch):
    package_modules = [x for x in sys.modules.keys() if "dioptra" in x]
    prefect_modules = [x for x in sys.modules.keys() if "prefect" in x]

    for module in package_modules + prefect_modules:
        monkeypatch.delitem(sys.modules, module)

    monkeypatch.setitem(sys.modules, "prefect", None)

    return (
        importlib.import_module("dioptra.pyplugs"),
        importlib.import_module("dioptra.sdk.exceptions"),
    )


def test_package_not_empty(plugin_package):
    """Test that names() finds some plugins in package"""
    plugins = pyplugs.names(plugin_package)
    assert len(plugins) > 0


def test_package_empty():
    """Test that names() does not find plugins in the dioptra.pyplugs
    package.
    """
    lib_plugins = pyplugs.names("dioptra.pyplugs")
    assert len(lib_plugins) == 0


def test_list_funcs(plugin_package):
    """Test that funcs() finds some plugins in package"""
    plugins = pyplugs.funcs(plugin_package, "plugin_parts")
    assert len(plugins) == 3


def test_package_non_existing():
    """Test that a nonexistent package raises an appropriate error"""
    with pytest.raises(UnknownPackageError):
        pyplugs.names("dioptra.pyplugs.non_existent")


def test_plugin_exists(plugin_package):
    """Test that an existing plugin returns its own plugin name"""
    plugin_name = pyplugs.names(plugin_package)[0]
    assert pyplugs.info(plugin_package, plugin_name).plugin_name == plugin_name


@pytest.mark.parametrize("plugin_name", ["no_plugins", "non_existent"])
def test_plugin_not_exists(plugin_package, plugin_name):
    """Test that a non-existing plugin raises UnknownPluginError

    Tests both for an existing module (no_plugins) and a nonexistent module
    (non_existent).
    """
    with pytest.raises(UnknownPluginError):
        pyplugs.info(plugin_package, plugin_name)


@pytest.mark.parametrize("task_func_name", ["get_task", "call_task"])
def test_prefect_dependency_not_installed(
    pyplugs_no_prefect, plugin_package, task_func_name
):
    plugin_name = "plugin_plain"
    pyplugs, exceptions = pyplugs_no_prefect
    task_func = getattr(pyplugs, task_func_name)

    with pytest.raises(exceptions.PrefectDependencyError):
        task_func(plugin_package, plugin=plugin_name)


@pytest.mark.parametrize("task_factory_name", ["get_task_factory", "call_task_factory"])
def test_factory_prefect_dependency_not_installed(
    pyplugs_no_prefect, plugin_package, task_factory_name
):
    pyplugs, exceptions = pyplugs_no_prefect
    task_factory_func = getattr(pyplugs, task_factory_name)

    with pytest.raises(exceptions.PrefectDependencyError):
        task_factory_func(plugin_package)


def test_exists(plugin_package):
    """Test that exists() function correctly identifies existing plugins"""
    assert pyplugs.exists(plugin_package, "plugin_parts") is True
    assert pyplugs.exists(plugin_package, "no_plugins") is False
    assert pyplugs.exists(plugin_package, "non_existent") is False


def test_exists_on_non_existing_package():
    """Test that exists() correctly returns False for non-existing packages"""
    assert pyplugs.exists("non_existent_package", "plugin_parts") is False
    assert pyplugs.exists("non_existent_package", "non_existent") is False


def test_call_existing_plugin(plugin_package):
    """Test that calling a test-plugin works, and returns a string"""
    plugin_name = pyplugs.names(plugin_package)[0]
    return_value = pyplugs.call(plugin_package, plugin_name)
    assert isinstance(return_value, str)


def test_call_non_existing_plugin():
    """Test that calling a non-existing plugin raises an error"""
    with pytest.raises(UnknownPluginError):
        pyplugs.call("dioptra.pyplugs", "non_existent")


def test_ordered_plugin(plugin_package):
    """Test that order of plugins can be customized"""
    plugin_names = pyplugs.names(plugin_package)
    assert plugin_names[0] == "plugin_first"
    assert plugin_names[-1] == "plugin_last"


def test_default_part(plugin_package):
    """Test that first registered function in a plugin is called by default"""
    plugin_name = "plugin_parts"
    default = pyplugs.call(plugin_package, plugin_name)
    explicit = pyplugs.call(plugin_package, plugin_name, func="plugin_default")
    assert default == explicit


def test_call_non_existing_func(plugin_package):
    """Test that calling a non-existing plug-in function raises an error"""
    plugin_name = "plugin_parts"
    func_name = "non_existent"
    with pytest.raises(UnknownPluginFunctionError):
        pyplugs.call(plugin_package, plugin_name, func=func_name)


def test_short_doc(plugin_package):
    """Test that we can retrieve the short docstring from a plugin"""
    plugin_name = "plugin_plain"
    doc = pyplugs.info(plugin_package, plugin_name).description
    assert doc == "A plain plugin"


def test_long_doc(plugin_package):
    """Test that we can retrieve the long docstring from a plugin"""
    plugin_name = "plugin_plain"
    doc = pyplugs.info(plugin_package, plugin_name).doc
    assert doc == "This is the plain docstring."


def test_names_factory(plugin_package):
    """Test that the names factory can retrieve names in package"""
    names = pyplugs.names_factory(plugin_package)
    factory_names = names()
    pyplugs_names = pyplugs.names(plugin_package)
    assert factory_names == pyplugs_names


def test_funcs_factory(plugin_package):
    """Test that the funcs factory can retrieve funcs within plugin"""
    plugin_name = "plugin_parts"
    funcs = pyplugs.funcs_factory(plugin_package)
    factory_funcs = funcs(plugin_name)
    pyplugs_funcs = pyplugs.funcs(plugin_package, plugin_name)
    assert factory_funcs == pyplugs_funcs


def test_info_factory(plugin_package):
    """Test that the info factory can retrieve info in package"""
    plugin_name = "plugin_parts"
    info = pyplugs.info_factory(plugin_package)
    factory_info = info(plugin_name)
    pyplugs_info = pyplugs.info(plugin_package, plugin=plugin_name)
    assert factory_info == pyplugs_info


def test_exists_factory(plugin_package):
    """Test that the exists factory can check for plugins in a package"""
    exists = pyplugs.exists_factory(plugin_package)
    assert exists("plugin_parts") is True
    assert exists("no_plugins") is False
    assert exists("non_existent") is False


def test_get_factory(plugin_package):
    """Test that the get factory can retrieve get in package"""
    plugin_name = "plugin_parts"
    get = pyplugs.get_factory(plugin_package)
    factory_get = get(plugin_name)
    pyplugs_get = pyplugs.get(plugin_package, plugin=plugin_name)
    assert factory_get == pyplugs_get


def test_call_factory(plugin_package):
    """Test that the call factory can retrieve call in package"""
    plugin_name = "plugin_parts"
    call = pyplugs.call_factory(plugin_package)
    factory_call = call(plugin_name)
    pyplugs_call = pyplugs.call(plugin_package, plugin=plugin_name)
    assert factory_call == pyplugs_call


def test_get_task_factory(plugin_package):
    """Test that the get task factory can retrieve a task get in package"""
    plugin_name = "plugin_parts"
    get_task = pyplugs.get_task_factory(plugin_package)

    with Flow("Test Get Task Factory") as flow:  # noqa: F841
        factory_get_task = get_task(plugin_name)
        pyplugs_get_task = pyplugs.get_task(plugin_package, plugin=plugin_name)
        assert factory_get_task.is_equal(pyplugs_get_task)


def test_call_task_factory(plugin_package):
    """Test that the call task factory can retrieve a task call in package"""
    plugin_name = "plugin_parts"
    call_task = pyplugs.call_task_factory(plugin_package)

    with Flow("Test Call Task Factory") as flow:
        factory_call_task = call_task(plugin_name)
        pyplugs_call_task = pyplugs.call_task(plugin_package, plugin=plugin_name)

    state = flow.run()
    assert (
        state.result[factory_call_task].result == state.result[pyplugs_call_task].result
    )


def test_call_tasks_with_nout(plugin_package):
    """Test that the task_nout decorator handles multiple argument returns in prefect
    tasks.
    """
    plugin_name = "plugin_task_nout"
    call_task = pyplugs.call_task_factory(plugin_package)

    with Flow("Test Call Task with nout") as flow:
        factory_result_1, factory_result_2 = call_task(
            plugin_name, func="plugin_with_nout"
        )
        pyplugs_result_1, pyplugs_result_2 = pyplugs.call_task(
            plugin_package, plugin=plugin_name, func="plugin_with_nout"
        )

    state = flow.run()
    assert (
        state.result[factory_result_1].result == state.result[pyplugs_result_1].result
    )
    assert (
        state.result[factory_result_2].result == state.result[pyplugs_result_2].result
    )


def test_call_tasks_without_nout(plugin_package):
    """Test that omitting the task_nout decorator is not compatible with multiple
    argument returns in prefect tasks.
    """
    plugin_name = "plugin_task_nout"

    with pytest.raises(TypeError):
        with Flow("Test Call Task without nout") as flow:
            result_1, result_2 = pyplugs.call_task(
                plugin_package, plugin=plugin_name, func="plugin_without_nout"
            )

            _ = flow.run()



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Shared utilities for REST API unit tests.

This module contains shared functionality used across test suites for each of the REST
API endpoints.
"""
import datetime
from typing import Any


def is_iso_format(date_string: str) -> bool:
    try:
        datetime.datetime.fromisoformat(date_string)
        return True

    except ValueError:
        return False


def is_timeout_format(timeout_string: str) -> bool:
    try:
        parsed_timeout_string = timeout_string.split("h")
        assert len(parsed_timeout_string) == 2
        assert parsed_timeout_string[-1] == ""
        assert parsed_timeout_string[0].isnumeric()
        return True

    except ValueError:
        return False


def assert_user_ref_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    assert isinstance(response["user"]["id"], int)
    assert isinstance(response["user"]["username"], str)
    assert isinstance(response["user"]["url"], str)
    assert response["user"]["id"] == expected_contents["user_id"]


def assert_group_ref_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    assert isinstance(response["group"]["id"], int)
    assert isinstance(response["group"]["name"], str)
    assert isinstance(response["group"]["url"], str)
    assert response["group"]["id"] == expected_contents["group_id"]


def assert_tag_ref_contents_matches_expectations(response: dict[str, Any]) -> None:
    for tag in response["tags"]:
        assert isinstance(tag["id"], int)
        assert isinstance(tag["name"], str)
        assert isinstance(tag["url"], str)


def assert_base_resource_ref_contents_matches_expectations(
    resource_type: str,
    response: dict[str, Any],
    expected_contents: dict[str, Any],
) -> None:
    assert isinstance(response[resource_type]["id"], int)
    assert isinstance(response[resource_type]["snapshotId"], int)
    assert isinstance(response[resource_type]["url"], str)
    assert_group_ref_contents_matches_expectations(
        response=response[resource_type], expected_contents=expected_contents
    )


def assert_queue_ref_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    assert isinstance(response["queue"]["name"], str)
    assert_base_resource_ref_contents_matches_expectations(
        resource_type="queue", response=response, expected_contents=expected_contents
    )
    assert response["queue"]["id"] == expected_contents["queue_id"]


def assert_experiment_ref_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    assert isinstance(response["experiment"]["name"], str)
    assert_base_resource_ref_contents_matches_expectations(
        resource_type="experiment",
        response=response,
        expected_contents=expected_contents,
    )
    assert response["experiment"]["id"] == expected_contents["experiment_id"]


def assert_entrypoint_ref_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    assert isinstance(response["entrypoint"]["name"], str)
    assert_base_resource_ref_contents_matches_expectations(
        resource_type="entrypoint",
        response=response,
        expected_contents=expected_contents,
    )
    assert response["entrypoint"]["id"] == expected_contents["entrypoint_id"]



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import os
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Iterable, Iterator, List

ENV_VAR_PLUGIN_DIR = "DIOPTRA_PLUGIN_DIR"


@contextmanager
def plugin_dirs(dirs: Iterable[str] = (".",)) -> Iterator[None]:
    sys_path_copy: List[str] = sys.path.copy()
    env_var_dirs: List[str] = [
        str(Path(x).resolve())
        for x in os.getenv(ENV_VAR_PLUGIN_DIR, default="").split(":")
        if x
        and Path(x).is_absolute()
        and Path(x).resolve().is_dir()
        and str(Path(x).resolve()) not in sys.path
    ]
    dirs = [
        str(Path(x).resolve())
        for x in dirs
        if x and Path(x).resolve().is_dir() and str(Path(x).resolve()) not in sys.path
    ]

    try:
        sys.path = _remove_duplicate_paths(
            [*sys.path[:1], *env_var_dirs, *dirs, *sys.path[1:]]
        )
        yield

    finally:
        sys.path = sys_path_copy


def _remove_duplicate_paths(dirs: Iterable[str]) -> List[str]:
    return list(dict.fromkeys(dirs))



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode

"""Test suite for experiment operations.

This module contains a set of tests that validate the supported CRUD operations and
additional functionalities for the experiment entity. The tests ensure that the
experiments can be registered, retrieved, and deleted as expected through the REST API.
"""

from typing import Any

import pytest
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy
from werkzeug.test import TestResponse

from dioptra.restapi.routes import V1_EXPERIMENTS_ROUTE, V1_ROOT

from ..lib import actions, asserts, helpers

# -- Actions ---------------------------------------------------------------------------


def modify_experiment(
    client: FlaskClient,
    experiment_id: int,
    new_name: str,
    new_entrypoints: list[int],
    new_description: str,
) -> TestResponse:
    """Modify an experiment using the API.

    Args:
        client: The Flask test client.
        experiment_id: The id of the experiment to modify.
        new_name: The new name to assign to the experiment.
        new_entrypoints: The new entrypoints to associate with the experiment.
        new_description: The new description to assign to the experiment.

    Returns:
        The response from the API.
    """

    payload = {
        "name": new_name,
        "entrypoints": new_entrypoints,
        "description": new_description,
    }

    return client.put(
        f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}/{experiment_id}",
        json=payload,
        follow_redirects=True,
    )


def delete_experiment_with_id(client: FlaskClient, experiment_id: int) -> TestResponse:
    """Delete an experiment using the API.

    Args:
        client: The Flask test client.
        experiment_id: The id of the experiment to delete.

    Returns:
        The response from the API.
    """
    return client.delete(
        f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}/{experiment_id}",
        follow_redirects=True,
    )


# -- Assertions ------------------------------------------------------------------------


def assert_experiment_response_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    """Assert that the experiment response is valid.

    Args:
        response: The actual response from the API.
        expected_contents: The expected response from the API.

    Raises:
        AssertionError: If the API response does not match the expected response or
        if the response contents are not valid.
    """
    expected_keys = {
        "id",
        "snapshot",
        "group",
        "user",
        "createdOn",
        "lastModifiedOn",
        "latestSnapshot",
        "hasDraft",
        "name",
        "entrypoints",
        "description",
        "tags",
    }
    assert set(response.keys()) == expected_keys

    # Validate the non-Ref fields
    assert isinstance(response["id"], int)
    assert isinstance(response["snapshot"], int)
    assert isinstance(response["name"], str)
    assert isinstance(response["description"], str)
    assert isinstance(response["createdOn"], str)
    assert isinstance(response["lastModifiedOn"], str)
    assert isinstance(response["latestSnapshot"], bool)

    assert response["name"] == expected_contents["name"]
    assert response["description"] == expected_contents["description"]

    assert helpers.is_iso_format(response["createdOn"])
    assert helpers.is_iso_format(response["lastModifiedOn"])

    # Validate the UserRef structure
    assert isinstance(response["user"]["id"], int)
    assert isinstance(response["user"]["username"], str)
    assert isinstance(response["user"]["url"], str)
    assert response["user"]["id"] == expected_contents["user_id"]

    # Validate the GroupRef structure
    assert isinstance(response["group"]["id"], int)
    assert isinstance(response["group"]["name"], str)
    assert isinstance(response["group"]["url"], str)
    assert response["group"]["id"] == expected_contents["group_id"]

    # Validate the EntryPointRef structure
    for entrypoint in response["entrypoints"]:
        assert isinstance(entrypoint["id"], int)
        assert isinstance(entrypoint["url"], str)
        assert isinstance(entrypoint["name"], str)
        assert isinstance(entrypoint["group"]["id"], int)
        assert isinstance(entrypoint["group"]["url"], str)
        assert isinstance(entrypoint["group"]["name"], str)

    # Validate the TagRef structure
    for tag in response["tags"]:
        assert isinstance(tag["id"], int)
        assert isinstance(tag["name"], str)
        assert isinstance(tag["url"], str)


def assert_retrieving_experiment_by_id_works(
    client: FlaskClient,
    experiment_id: int,
    expected: dict[str, Any],
) -> None:
    """Assert that retrieving an experiment by id works.

    Args:
        client: The Flask test client.
        experiment_id: The id of the experiment to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}/{experiment_id}", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_experiment_entrypoints_matches_expected_entrypoints(
    client: FlaskClient,
    experiment_id: int,
    expected_entrypoints: dict[str, Any],
) -> None:
    """Assert that an experiment's associated entrypoints can be updated for an
    experiment with specified id.

    Args:
        client: The Flask test client.
        experiment_id: The id of the experiment to add entrypoints.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}/{experiment_id}",
        follow_redirects=True,
    )
    assert response.status_code == 200
    response_entrypoints = [
        entrypoint["id"] for entrypoint in response.get_json()["entrypoints"]
    ]
    assert response_entrypoints == expected_entrypoints


def assert_retrieving_all_experiments_works(
    client: FlaskClient,
    expected: list[dict[str, Any]],
    group_id: int | None = None,
    search: str | None = None,
    paging_info: dict[str, Any] | None = None,
) -> None:
    """Assert that retrieving all experiments works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.
        group_id: The group ID used in query parameters.
        search: The search string used in query parameters.
        paging_info: The paging information used in query parameters.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    if group_id is not None:
        query_string["groupId"] = group_id

    if search is not None:
        query_string["search"] = search

    if paging_info is not None:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_sorting_experiment_works(
    client: FlaskClient,
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Assert that experiments can be sorted by column ascending/descending.

    Args:
        client: The Flask test client.
        expected: The expected order of experiment ids after sorting.
            See test_experiment_sort for expected orders.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    query_string["sortBy"] = sortBy
    query_string["descending"] = descending

    response = client.get(
        f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )

    response_data = response.get_json()
    experiment_ids = [experiment["id"] for experiment in response_data["data"]]

    assert response.status_code == 200 and experiment_ids == expected


def assert_experiment_name_matches_expected_name(
    client: FlaskClient, experiment_id: int, expected_name: str
) -> None:
    """Assert that the name of an experiment matches the expected name.

    Args:
        client: The Flask test client.
        experiment_id: The id of the experiment to retrieve.
        expected_name: The expected name of the experiment.

    Raises:
        AssertionError: If the response status code is not 200 or if the name of the
            experiment does not match the expected name.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}/{experiment_id}",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["name"] == expected_name


def assert_experiment_is_not_found(
    client: FlaskClient,
    experiment_id: int,
) -> None:
    """Assert that an experiment is not found.

    Args:
        client: The Flask test client.
        experiment_id: The id of the experiment to retrieve.

    Raises:
        AssertionError: If the response status code is not 404.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}/{experiment_id}",
        follow_redirects=True,
    )
    assert response.status_code == 404


# -- Tests -----------------------------------------------------------------------------


def test_create_experiment(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> None:
    """Test that experiments can be registered and retrieved using the API.

    Given an authenticated user, this test validates the following sequence of actions:

    - A user registers an experiment named "experiment1".
    - The response is valid and matches the expected values of the registration request.
    - The user is able to retrieve information about the experiment using the
      experiment id.
    """

    name = "experiment1"
    user_id = auth_account["id"]
    group_id = auth_account["default_group_id"]
    description = "test description"

    experiment1_response = actions.register_experiment(
        client,
        name=name,
        group_id=group_id,
        description=description,
    )
    experiment1_expected = experiment1_response.get_json()
    assert_experiment_response_contents_matches_expectations(
        response=experiment1_expected,
        expected_contents={
            "name": name,
            "user_id": user_id,
            "group_id": group_id,
            "description": description,
        },
    )
    assert_retrieving_experiment_by_id_works(
        client,
        experiment_id=experiment1_expected["id"],
        expected=experiment1_expected,
    )


def test_experiment_get_all(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_experiments: dict[str, Any],
) -> None:
    """Test that all experiments can be retrieved.

    Given an authenticated user and registered experiments, this test validates the
        following sequence of actions:

    - A user registers three experiments, "experiment1", "experiment2", and
      "experiment3".
    - The user is able to retrieve a list of all registered experiments.
    - The returned list of experiments matches the full list of registered experiments.
    """
    experiment1_expected = registered_experiments["experiment1"]
    experiment2_expected = registered_experiments["experiment2"]
    experiment3_expected = registered_experiments["experiment3"]
    experiment_expected_list = [
        experiment1_expected,
        experiment2_expected,
        experiment3_expected,
    ]

    assert_retrieving_all_experiments_works(client, expected=experiment_expected_list)


@pytest.mark.parametrize(
    "sortBy, descending , expected",
    [
        (None, None, ["experiment1", "experiment2", "experiment3"]),
        ("name", True, ["experiment3", "experiment2", "experiment1"]),
        ("name", False, ["experiment1", "experiment2", "experiment3"]),
        ("createdOn", True, ["experiment3", "experiment2", "experiment1"]),
        ("createdOn", False, ["experiment1", "experiment2", "experiment3"]),
    ],
)
def test_experiment_sort(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_experiments: dict[str, Any],
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Test that experiments can be sorted by column.

    Given an authenticated user and registered experiments, this test validates the
      following sequence of actions:

    - A user registers three experiments: "experiment1", "experiment2", "experiment3".
    - The user is able to retrieve a list of all registered experiments sorted by a
      column ascending/descending.
    - The returned list of experiments matches the order in the parametrize lists above.
    """

    expected_ids = [
        registered_experiments[expected_name]["id"] for expected_name in expected
    ]
    assert_sorting_experiment_works(client, sortBy, descending, expected=expected_ids)


def test_experiment_search_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_experiments: dict[str, Any],
) -> None:
    """Test that experiments can be queried with a search term.

    Given an authenticated user and registered experiments, this test validates the
        following sequence of actions:

    - The user is able to retrieve a list of all registered experiments with a name
      that contains 'experiment'.
    - The returned list of experiments matches the expected matches from the query.
    """
    experiment1_expected = registered_experiments["experiment1"]
    experiment2_expected = registered_experiments["experiment2"]
    experiment3_expected = registered_experiments["experiment3"]
    experiment_expected_list = [
        experiment1_expected,
        experiment2_expected,
        experiment3_expected,
    ]

    assert_retrieving_all_experiments_works(
        client,
        expected=experiment_expected_list,
        search="description:*description*",
    )

    assert_retrieving_all_experiments_works(
        client,
        expected=experiment_expected_list,
        search="name:*experiment*",
    )

    assert_retrieving_all_experiments_works(
        client,
        expected=experiment_expected_list,
        search="*",
    )


def test_experiment_group_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_experiments: dict[str, Any],
) -> None:
    """Test that experiments can be retrieved using a group filter.

    Given an authenticated user and registered experiments, this test validates the
        following sequence of actions:

    - The user is able to retrieve a list of all registered experiments that are owned
      by the default group.
    - The returned list of experiments matches the expected list owned by the default
      group.
    """
    experiment1_expected = registered_experiments["experiment1"]
    experiment2_expected = registered_experiments["experiment2"]
    experiment3_expected = registered_experiments["experiment3"]
    experiment_expected_list = [
        experiment1_expected,
        experiment2_expected,
        experiment3_expected,
    ]

    assert_retrieving_all_experiments_works(
        client,
        expected=experiment_expected_list,
        group_id=auth_account["default_group_id"],
    )


def test_experiment_get_by_id(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_experiments: dict[str, Any],
) -> None:
    """Test that experiments can be retrieved by their unique ID.

    Given an authenticated user and registered experiments, this test validates the
        following sequence of actions:

    - The user is able to retrieve a single experiment by its ID.
    - The response is a single experiment with a matching ID.
    """
    experiment2_expected = registered_experiments["experiment2"]

    assert_retrieving_experiment_by_id_works(
        client,
        experiment_id=experiment2_expected["id"],
        expected=experiment2_expected,
    )


def test_rename_experiment(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_experiments: dict[str, Any],
) -> None:
    """Test that an experiment can be renamed.

    Given an authenticated user and registered experiments, this test validates the
    following sequence of actions:

    - The user issues a request to change the name of an experiment.
    - The user retrieves information about the same experiment and it reflects
      the name change.
    """
    updated_experiment_name = "experiment0"
    experiment_to_rename = registered_experiments["experiment1"]
    existing_description = experiment_to_rename["description"]
    modify_experiment(
        client,
        experiment_id=experiment_to_rename["id"],
        new_name=updated_experiment_name,
        new_description=existing_description,
        new_entrypoints=[],  # TODO use existing entrypoints
    )
    assert_experiment_name_matches_expected_name(
        client,
        experiment_id=experiment_to_rename["id"],
        expected_name=updated_experiment_name,
    )


def test_delete_experiment_by_id(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_experiments: dict[str, Any],
) -> None:
    """Test that an experiment can be deleted by referencing its id.

    Given an authenticated user and registered experiments, this test validates the
        following sequence of actions:

    - The user deletes the "experiment3" experiment by referencing its id.
    - The user attempts to retrieve information about the "experiment3" experiment,
      and the response indicates it is no longer found.
    """
    experiment_to_delete = registered_experiments["experiment3"]
    delete_experiment_with_id(client, experiment_id=experiment_to_delete["id"])
    assert_experiment_is_not_found(client, experiment_id=experiment_to_delete["id"])


def test_experiment_get_entrypoints(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_entrypoints: dict[str, Any],
) -> None:
    """Test that an experiment's associated entrypoints can be retrieved by referencing
    the experiment's id.

    This test validates the following sequence of actions:

    - A user registers an experiment named "experiment1".
    - The user is able to retrieve entrypoints information about the "experiment1"
      experiment that matches the information provided during registration.
    """
    name = "experiment1"
    user_id = auth_account["id"]
    group_id = auth_account["default_group_id"]
    description = "test description"
    entrypoints = [registered_entrypoints["entrypoint1"]["id"]]
    registration_response = actions.register_experiment(
        client,
        name=name,
        group_id=group_id,
        entrypoint_ids=entrypoints,
        description=description,
    )
    experiment_json = registration_response.get_json()
    experiment_id = experiment_json["id"]

    assert_experiment_response_contents_matches_expectations(
        response=experiment_json,
        expected_contents={
            "name": name,
            "user_id": user_id,
            "group_id": group_id,
            "entrypoint_ids": entrypoints,
            "description": description,
        },
    )

    assert_experiment_entrypoints_matches_expected_entrypoints(
        client,
        experiment_id=experiment_id,
        expected_entrypoints=entrypoints,
    )


@pytest.mark.skip(reason="entrypoints not yet implemented")
# @pytest.mark.parametrize(
#     "initial_entrypoints",
#     "new_entrypoints",
#     "expected_entrypoints",
#     [
#         ([1, 2, 3], [4], [1, 2, 3, 4]),
#         ([], [1, 2], [1, 2]),
#         ([1, 2, 3], [2], [1, 2, 3]),
#     ]
# )
def test_experiment_add_entrypoints(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    initial_entrypoints: list[int],
    new_entrypoints: list[int],
    expected_entrypoints: list[int],
) -> None:
    """Test that entrypoints can be added to and retrieved from an experiment using
    the API.

    Given an authenticated user, this test validates the following sequence of actions:

    - A user creates an experiment, "experiment1" with initial entrypoints.
    - The user is able to add an associated entry point to "experiment1".
    - The user is able to retrieve the same experiment's information, and it reflects
      the new entrypoint association.
    """
    name = "experiment1"
    description = "test description"
    user_id = auth_account["id"]
    group_id = auth_account["default_group_id"]

    experiment1_response = actions.register_experiment(
        client,
        name=name,
        group_id=group_id,
        entrypoints=initial_entrypoints,
        description=description,
    )
    experiment1_json = experiment1_response.get_json()
    experiment_id = experiment1_json["id"]
    assert_experiment_response_contents_matches_expectations(
        response=experiment1_json,
        expected_contents={
            "name": name,
            "user_id": user_id,
            "group_id": group_id,
            "entrypoints": initial_entrypoints,
            "description": description,
        },
    )

    modify_experiment(
        client,
        experiment_id=experiment1_json["id"],
        new_name=name,
        new_description=description,
        new_entrypoints=initial_entrypoints + new_entrypoints,
    )

    assert_experiment_entrypoints_matches_expected_entrypoints(
        client, experiment_id=experiment_id, expected_entrypoints=expected_entrypoints
    )


@pytest.mark.skip(reason="entrypoints not yet implemented")
# @pytest.mark.parametrize(
#     "initial_entrypoints", "new_entrypoints", "expected_entrypoints",
#     [
#         ([1, 2, 3], [4, 5], [4, 5]),
#         ([1, 2, 3], [2, 3, 4], [2, 3, 4]),
#         ([1, 2, 3], [], []),
#     ],
# )
def test_experiment_modify_entrypoints(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    initial_entrypoints: list[int],
    new_entrypoints: list[int],
    expected_entrypoints: list[int],
) -> None:
    """Test that entrypoints can be replaced and retrieved from an experiment using
    the API.

    Given an authenticated user, this test validates the following sequence of actions:

    - A user creates an experiment, "experiment1".
    - The user is able to replace associated entrypoints for "experiment1".
    - The user is able to retrieve the same experiment's information, and it reflects
      the new entrypoint associations.
    """
    name = "experiment1"
    description = "test description"
    user_id = auth_account["id"]
    group_id = auth_account["default_group_id"]

    experiment1_response = actions.register_experiment(
        client,
        name=name,
        group_id=group_id,
        entrypoints=initial_entrypoints,
        description=description,
    )
    experiment1_json = experiment1_response.get_json()
    experiment_id = experiment1_json["id"]
    assert_experiment_response_contents_matches_expectations(
        response=experiment1_json,
        expected_contents={
            "name": name,
            "user_id": user_id,
            "group_id": group_id,
            "entrypoints": initial_entrypoints,
            "description": description,
        },
    )
    modify_experiment(
        client,
        experiment_id=experiment_id,
        new_name=name,
        new_entrypoints=new_entrypoints,
        new_description=description,
    )

    assert_experiment_entrypoints_matches_expected_entrypoints(
        client,
        experiment_id=experiment_id,
        expected_entrypoints=expected_entrypoints,
    )


@pytest.mark.skip(reason="entrypoints not yet implemented")
# @pytest.mark.parametrize(
#     "initial_entrypoints", "entrypoints_to_remove", "expected_entrypoints",
#     [
#         ([1, 2, 3], [2, 3], [3]),
#         ([1, 2, 3], [4], [1, 2, 3]),
#         ([], [1], []),
#     ]
# )
def test_experiment_remove_entrypoints(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    initial_entrypoints: list[int],
    entrypoints_to_remove: list[int],
    expected_entrypoints: list[int],
) -> None:
    """Test that entrypoints can be removed from an experiment using the API.

    Given an authenticated user, this test validates the following sequence of actions:

    - A user creates an experiment, "experiment1".
    - The user is able to remove associated entrypoints for "experiment1".
    - The user is able to retrieve the same experiment's information, and it reflects
      the new entrypoint associations.
    """
    name = "experiment1"
    description = "test description"
    user_id = auth_account["id"]
    group_id = auth_account["default_group_id"]

    experiment1_response = actions.register_experiment(
        client,
        name=name,
        group_id=group_id,
        entrypoints=initial_entrypoints,
        description=description,
    )
    experiment1_json = experiment1_response.get_json()
    experiment_id = experiment1_json["id"]
    assert_experiment_response_contents_matches_expectations(
        response=experiment1_json,
        expected_contents={
            "name": name,
            "user_id": user_id,
            "group_id": group_id,
            "entrypoints": initial_entrypoints,
            "description": description,
        },
    )

    updated_entrypoints = []
    for entry in initial_entrypoints:
        if entry not in entrypoints_to_remove:
            updated_entrypoints.append(entry)

    modify_experiment(
        client,
        experiment_id=experiment_id,
        new_name=name,
        new_entrypoints=updated_entrypoints,
        new_description=description,
    )

    assert_experiment_entrypoints_matches_expected_entrypoints(
        client,
        experiment_id=experiment_id,
        expected_entrypoints=expected_entrypoints,
    )


def test_manage_existing_experiment_draft(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_experiments: dict[str, Any],
) -> None:
    """Test that a draft of an existing experiment can be created and managed by the
        user

    Given an authenticated user and registered experiments, this test validates the
    following sequence of actions:

    - The user creates a draft of an existing experiment
    - The user retrieves information about the draft and gets the expected response
    - The user attempts to create another draft of the same existing experiment
    - The request fails with an appropriate error message and response code.
    - The user modifies the name of the experiment in the draft
    - The user retrieves information about the draft and gets the expected response
    - The user deletes the draft
    - The user attempts to retrieve information about the deleted draft.
    - The request fails with an appropriate error message and response code.
    """
    experiment = registered_experiments["experiment1"]
    name = "draft"
    new_name = "draft2"
    description = "description"

    # test creation
    payload = {"name": name, "description": description, "entrypoints": [1]}
    expected = {
        "user_id": auth_account["id"],
        "group_id": experiment["group"]["id"],
        "resource_id": experiment["id"],
        "resource_snapshot_id": experiment["snapshot"],
        "num_other_drafts": 0,
        "payload": payload,
    }
    response = actions.create_existing_resource_draft(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        resource_id=experiment["id"],
        payload=payload,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(response, expected)
    asserts.assert_retrieving_draft_by_resource_id_works(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        resource_id=experiment["id"],
        expected=response,
    )
    asserts.assert_creating_another_existing_draft_fails(
        client, resource_route=V1_EXPERIMENTS_ROUTE, resource_id=experiment["id"]
    )

    # test modification
    payload = {"name": new_name, "description": description, "entrypoints": [3, 2]}
    expected = {
        "user_id": auth_account["id"],
        "group_id": experiment["group"]["id"],
        "resource_id": experiment["id"],
        "resource_snapshot_id": experiment["snapshot"],
        "num_other_drafts": 0,
        "payload": payload,
    }
    response = actions.modify_existing_resource_draft(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        resource_id=experiment["id"],
        payload=payload,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(response, expected)

    # test deletion
    actions.delete_existing_resource_draft(
        client, resource_route=V1_EXPERIMENTS_ROUTE, resource_id=experiment["id"]
    )
    asserts.assert_existing_draft_is_not_found(
        client, resource_route=V1_EXPERIMENTS_ROUTE, resource_id=experiment["id"]
    )


def test_manage_new_experiment_drafts(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> None:
    """Test that drafts of experiment can be created and managed by the user

    Given an authenticated user, this test validates the following sequence of actions:

    - The user creates two experiment drafts
    - The user retrieves information about the drafts and gets the expected response
    - The user modifies the description of the experiment in the first draft
    - The user retrieves information about the draft and gets the expected response
    - The user deletes the first draft
    - The user attempts to retrieve information about the deleted draft.
    - The request fails with an appropriate error message and response code.
    """
    group_id = auth_account["groups"][0]["id"]
    drafts = {
        "draft1": {
            "name": "experiment1",
            "description": "my experiment",
            "entrypoints": [3],
        },
        "draft2": {"name": "experiment2", "description": None, "entrypoints": [3, 4]},
    }

    # test creation
    draft1_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": drafts["draft1"],
    }
    draft1_response = actions.create_new_resource_draft(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        group_id=group_id,
        payload=drafts["draft1"],
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        draft1_response, draft1_expected
    )
    asserts.assert_retrieving_draft_by_id_works(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        draft_id=draft1_response["id"],
        expected=draft1_response,
    )
    draft2_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": drafts["draft2"],
    }
    draft2_response = actions.create_new_resource_draft(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        group_id=group_id,
        payload=drafts["draft2"],
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        draft2_response, draft2_expected
    )
    asserts.assert_retrieving_draft_by_id_works(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        draft_id=draft2_response["id"],
        expected=draft2_response,
    )
    asserts.assert_retrieving_drafts_works(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        expected=[draft1_response, draft2_response],
    )

    # test modification
    draft1_mod = {"name": "draft1", "description": "new description", "entrypoints": []}
    draft1_mod_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": draft1_mod,
    }
    response = actions.modify_new_resource_draft(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        draft_id=draft1_response["id"],
        payload=draft1_mod,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        response, draft1_mod_expected
    )

    # test deletion
    actions.delete_new_resource_draft(
        client, resource_route=V1_EXPERIMENTS_ROUTE, draft_id=draft1_response["id"]
    )
    asserts.assert_new_draft_is_not_found(
        client, resource_route=V1_EXPERIMENTS_ROUTE, draft_id=draft1_response["id"]
    )


def test_manage_experiment_snapshots(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_experiments: dict[str, Any],
) -> None:
    """Test that different snapshots of an experiment can be retrieved by the user.

    Given an authenticated user and registered experiments, this test validates the
    following sequence of actions:

    - The user modifies an experiment
    - The user retrieves information about the original snapshot of the experiment and
      gets the expected response
    - The user retrieves information about the new snapshot of the experiment and gets
      the expected response
    - The user retrieves a list of all snapshots of the experiment and gets the expected
      response
    """
    experiment_to_rename = registered_experiments["experiment1"]
    modified_experiment = modify_experiment(
        client,
        experiment_id=experiment_to_rename["id"],
        new_name=experiment_to_rename["name"] + "modified",
        new_description=experiment_to_rename["description"],
        new_entrypoints=[],  # TODO use existing entrypoints
    ).get_json()
    modified_experiment.pop("hasDraft")
    modified_experiment.pop("entrypoints")
    experiment_to_rename.pop("hasDraft")
    experiment_to_rename.pop("entrypoints")
    experiment_to_rename["latestSnapshot"] = False
    experiment_to_rename["lastModifiedOn"] = modified_experiment["lastModifiedOn"]
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        resource_id=experiment_to_rename["id"],
        snapshot_id=experiment_to_rename["snapshot"],
        expected=experiment_to_rename,
    )
    expected_snapshots = [experiment_to_rename, modified_experiment]
    asserts.assert_retrieving_snapshots_works(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        resource_id=experiment_to_rename["id"],
        expected=expected_snapshots,
    )


def test_tag_experiment(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_experiments: dict[str, Any],
    registered_tags: dict[str, Any],
) -> None:
    """Test that tags can applied to experiments.

    Given an authenticated user and registered experiments, this test validates the
    following sequence of actions:

    """
    experiment = registered_experiments["experiment1"]
    tags = [tag["id"] for tag in registered_tags.values()]

    # test append
    response = actions.append_tags(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        resource_id=experiment["id"],
        tag_ids=[tags[0], tags[1]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1]]
    )
    response = actions.append_tags(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        resource_id=experiment["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1], tags[2]]
    )

    # test remove
    actions.remove_tag(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        resource_id=experiment["id"],
        tag_id=tags[1],
    )
    response = actions.get_tags(
        client, resource_route=V1_EXPERIMENTS_ROUTE, resource_id=experiment["id"]
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[2]]
    )

    # test modify
    response = actions.modify_tags(
        client,
        resource_route=V1_EXPERIMENTS_ROUTE,
        resource_id=experiment["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[1], tags[2]]
    )

    # test delete
    response = actions.remove_tags(
        client, resource_route=V1_EXPERIMENTS_ROUTE, resource_id=experiment["id"]
    )
    response = actions.get_tags(
        client, resource_route=V1_EXPERIMENTS_ROUTE, resource_id=experiment["id"]
    )
    asserts.assert_tags_response_contents_matches_expectations(response.get_json(), [])



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""A module of utility functions that are used across multiple endpoints.

.. |Namespace| replace:: :py:class:`~flask_restx.Namespace`
.. |ParametersSchema| replace:: :py:class:`~.ParametersSchema`
.. |RequestParser| replace:: :py:class:`~flask_restx.reqparse.RequestParser`
.. |Resource| replace:: :py:class:`~flask_restx.Resource`
"""
from __future__ import annotations

import datetime
import functools
from collections import Counter
from importlib.resources import as_file, files
from typing import Any, Callable, List, Protocol, Type, cast

from flask.views import View
from flask_restx import Api, Namespace, Resource, inputs
from flask_restx.reqparse import RequestParser
from injector import Injector
from marshmallow import Schema
from marshmallow import fields as ma
from marshmallow import missing
from marshmallow.schema import SchemaMeta
from typing_extensions import TypedDict
from werkzeug.datastructures import FileStorage

from dioptra.restapi.v1.shared.request_scope import set_request_scope_callbacks

from .custom_schema_fields import FileUpload


class ParametersSchema(TypedDict, total=False):
    """A schema of the parameters that can be passed to the |RequestParser|."""

    name: str
    type: object
    location: str
    required: bool
    default: Any | None
    help: str


def as_api_parser(
    api: Namespace, parameters_schema: List[ParametersSchema]
) -> RequestParser:
    """Creates a |RequestParser| from a list of |ParametersSchema| dictionaries.

    The primary use case for this utility function is to convert a submission form
    schema so that it can be passed to an `@api.expects` decorator and populate the
    form's fields in the Swagger documentation.

    Args:
        api: An endpoint's |Namespace| used for grouping |Resource|s.
        parameters_schema: A list of |ParametersSchema| dictionaries to add as arguments
            to `api.parser()`.

    Returns:
        A |RequestParser| object.
    """
    parser: RequestParser = api.parser()

    for form_kwargs in parameters_schema:
        parser.add_argument(**form_kwargs)

    return parser


def as_parameters_schema_list(
    schema: Schema | SchemaMeta, operation: str, location: str
) -> list[ParametersSchema]:
    """Converts a Marshmallow Schema into a list of ParametersSchema dictionaries.

    This primary use case for this function is for converting a Marshmallow Schema into
    a format that can passed to the `as_api_parser` utility function so that the Swagger
    documentation is correct.

    Args:
        schema: A Marshmallow Schema or SchemaMeta object.
        operation: The Schema operation the REST API is going to perform. Use "load"
            for a request and "dump" for a response.
        location: The location where the request or response parameters are stored.
            Must be either "args", "form", or "json".

    Returns:
        A list of ParametersSchema dictionaries.

    Raises:
        ValueError: If `operation` is not "dump" or "load".

    Note:
        This is a necessary workaround because the @accepts decorator ignores the
        load_only and dump_only settings on a Marshmallow Schema that's passed to the
        form_schema= argument when generating the Swagger documentation.
    """
    if operation not in {"dump", "load"}:
        raise ValueError(f"Invalid operation: {operation}. Must be 'dump' or 'load'")

    if isinstance(schema, SchemaMeta):
        schema = cast(Schema, schema())

    parameters_schema_list: list[ParametersSchema] = []
    for field in schema.fields.values():
        if operation == "dump" and not field.load_only:
            parameters_schema_list.append(
                create_parameters_schema(field, operation, location)
            )

        if operation == "load" and not field.dump_only:
            parameters_schema_list.append(
                create_parameters_schema(field, operation, location)
            )

    return parameters_schema_list


def create_parameters_schema(
    field: ma.Field, operation: str, location: str
) -> ParametersSchema:
    """Converts a Marshmallow Field into a ParametersSchema dictionary.

    Args:
        field: A Marshmallow Field object.
        operation: The Schema operation the REST API is going to perform. Use "load"
            for a request and "dump" for a response.
        location: The location where the request or response parameters are stored.
            Must be either "args", "form", or "json".

    Returns:
        A ParametersSchema dictionary.
    """
    parameter_type = TYPE_MAP_MA_TO_REQPARSE[type(field)]

    if parameter_type is FileStorage:
        location = "files"

    parameters_schema = ParametersSchema(
        name=cast(str, field.name),
        type=parameter_type,
        location=location,
        required=field.required,
    )

    if operation == "load" and field.load_default is not missing:
        parameters_schema["default"] = field.load_default

    if operation == "dump" and field.dump_default is not missing:
        parameters_schema["default"] = field.dump_default

    if field.metadata.get("description") is not None:
        parameters_schema["help"] = field.metadata["description"]

    return parameters_schema


def slugify(text: str) -> str:
    """Slugify a string.

    A "slugified" string is a string that has been lowercased and had all of its
    spaces replaced with hyphens.

    Args:
        text: The text to slugifiy.

    Returns:
        A slugified string
    """

    return text.lower().strip().replace(" ", "-")


def read_text_file(package: str, filename: str) -> str:
    """Read a text file from a specified package into a string.

    Args:
        package: The name of the Python package containing the text file. This should
            be a string representing the package's import path, for example
            "my_package.subpackage".
        filename: The base name of the text file, including its extension. For
            example, "data.txt".

    Returns:
        A string with the contents of the text file.
    """
    traversable = files(package).joinpath(filename)
    with as_file(traversable) as fp:
        return fp.read_text()


class _ClassBasedViewFunction(Protocol):
    """
    We distinguish a class-based view function from other view functions
    by looking for a "view_class" attribute on the function.
    """

    view_class: Type[View]

    def __call__(self, *args, **kwargs) -> Any: ...  # noqa: E704


def _new_class_view_function(
    func: _ClassBasedViewFunction, injector: Injector, api: Api
) -> Callable[..., Any]:
    """
    Create a view function which supports injection, based on the given
    class-based view function.  "Wrapping" func won't work here, in the sense
    that our view function can't delegate to func since the latter does not
    support dependency-injected view object creation.  So we create a brand new
    one (@wrap'd, so it has the look of func at least), which does
    dependency-injected view creation.

    Args:
        func: The old class-based view function
        injector: An injector
        api: The flask_restx Api instance

    Returns:
        A new view function
    """

    is_restx_resource = issubclass(func.view_class, Resource)

    additional_kwargs = {}
    if is_restx_resource:
        additional_kwargs["api"] = api

    # Honoring init_every_request is simple enough to do, so why not.
    # It was added in Flask 2.2.0; it behaved as though True, previously.
    init_every_request = getattr(func.view_class, "init_every_request", True)

    if not init_every_request:
        view_obj = injector.create_object(
            func.view_class, additional_kwargs=additional_kwargs
        )

    @functools.wraps(
        func,
        assigned=functools.WRAPPER_ASSIGNMENTS
        + ("view_class", "methods", "provide_automatic_options"),
    )
    def new_view_func(*args, **kwargs):
        nonlocal view_obj
        if init_every_request:
            view_obj = injector.create_object(
                func.view_class, additional_kwargs=additional_kwargs
            )

        return view_obj.dispatch_request(*args, **kwargs)

    if is_restx_resource:
        new_view_func = api.output(new_view_func)

    return new_view_func


def setup_injection(api: Api, injector: Injector) -> None:
    """
    Fixup the given flask app such that class-based view functions support
    dependency injection.

    Args:
        api: A flask_restx Api object.  This contains the flask app, and is
            also necessary to make restx views (resources) work with
            dependency injection.
        injector: An injector
    """

    new_view_func: Callable[..., Any]

    for key, func in api.app.view_functions.items():
        if hasattr(func, "view_class"):
            new_view_func = _new_class_view_function(func, injector, api)
            api.app.view_functions[key] = new_view_func

    set_request_scope_callbacks(api.app, injector)

    # Uncomment to see more detailed logging regarding dependency injection
    # in debug mode.
    # if api.app.debug:
    #     injector_logger = logging.getLogger("injector")
    #     injector_logger.setLevel(logging.DEBUG)


TYPE_MAP_MA_TO_REQPARSE = {
    ma.AwareDateTime: str,
    ma.Bool: inputs.boolean,
    ma.Boolean: inputs.boolean,
    ma.Constant: str,
    ma.Date: inputs.date_from_iso8601,
    ma.DateTime: inputs.datetime_from_iso8601,
    ma.Decimal: float,
    ma.Dict: dict,
    ma.Email: str,
    FileUpload: FileStorage,
    ma.Float: float,
    ma.Function: str,
    ma.Int: int,
    ma.Integer: int,
    ma.Length: float,
    ma.List: list,
    ma.Mapping: dict,
    ma.Method: str,
    ma.NaiveDateTime: inputs.datetime_from_iso8601,
    ma.Number: float,
    ma.Pluck: str,
    ma.Raw: str,
    ma.Str: str,
    ma.String: str,
    ma.Time: datetime.time.fromisoformat,
    ma.Url: str,
    ma.URL: str,
    ma.UUID: str,
}


# Validation Functions
def find_non_unique(name: str, parameters: list[dict[str, Any]]) -> list[str]:
    """
    Finds all values of a key that are not unique in a list of dictionaries.
    Useful for checking that a provided input satisfies uniqueness constraints.

    Note that the key name must be in every dictionary of the provided list.

    Args:
        name: the name of the parameter to check
        parameters: the input parameters to check

    Returns:
        A list of all values that were provided more than once, or an empty list if all
        values of the key were unique
    """
    name_count: Counter = Counter()
    # this line fails if a parameter is missing a "name" value
    name_count.update([parameter[name] for parameter in parameters])
    # create a list of all name values that appear more than once
    return [key for key in name_count.keys() if name_count[key] > 1]



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Test suite for entrypoint operations.

This module contains a set of tests that validate the CRUD operations and additional
functionalities for the entrypoint entity. The tests ensure that the entrypoints can be
registered, renamed, deleted, and locked/unlocked as expected through the REST API.
"""
import textwrap
from typing import Any

import pytest
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy
from werkzeug.test import TestResponse

from dioptra.restapi.routes import V1_ENTRYPOINTS_ROUTE, V1_EXPERIMENTS_ROUTE, V1_ROOT

from ..lib import actions, asserts, helpers

# -- Actions ---------------------------------------------------------------------------


def modify_entrypoint(
    client: FlaskClient,
    entrypoint_id: int,
    new_name: str,
    new_description: str,
    new_task_graph: str,
    new_parameters: list[dict[str, Any]],
    new_queue_ids: list[int],
) -> TestResponse:
    """Rename a entrypoint using the API.

    Args:
        client: The Flask test client.
        entrypoint_id: The id of the entrypoint to rename.
        new_name: The new name to assign to the entrypoint.
        new_description: The new description to assign to the entrypoint.

    Returns:
        The response from the API.
    """
    payload: dict[str, Any] = {
        "name": new_name,
        "description": new_description,
        "taskGraph": new_task_graph,
        "parameters": new_parameters,
        "queues": new_queue_ids,
    }

    return client.put(
        f"/{V1_ROOT}/{V1_ENTRYPOINTS_ROUTE}/{entrypoint_id}",
        json=payload,
        follow_redirects=True,
    )


def delete_entrypoint(
    client: FlaskClient,
    entrypoint_id: int,
) -> TestResponse:
    """Delete a entrypoint using the API.

    Args:
        client: The Flask test client.
        entrypoint_id: The id of the entrypoint to delete.

    Returns:
        The response from the API.
    """

    return client.delete(
        f"/{V1_ROOT}/{V1_ENTRYPOINTS_ROUTE}/{entrypoint_id}",
        follow_redirects=True,
    )


# -- Assertions ------------------------------------------------------------------------


def assert_entrypoint_response_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    """Assert that entrypoint response contents is valid.

    Args:
        response: The actual response from the API.
        expected_contents: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response or if the response contents is not
            valid.
    """
    expected_keys = {
        "id",
        "snapshot",
        "group",
        "user",
        "createdOn",
        "lastModifiedOn",
        "latestSnapshot",
        "hasDraft",
        "name",
        "description",
        "taskGraph",
        "parameters",
        "plugins",
        "queues",
        "tags",
    }
    assert set(response.keys()) == expected_keys

    # Validate the non-Ref fields
    assert isinstance(response["id"], int)
    assert isinstance(response["snapshot"], int)
    assert isinstance(response["name"], str)
    assert isinstance(response["description"], str)
    assert isinstance(response["createdOn"], str)
    assert isinstance(response["lastModifiedOn"], str)
    assert isinstance(response["latestSnapshot"], bool)
    assert isinstance(response["hasDraft"], bool)

    assert response["name"] == expected_contents["name"]
    assert response["description"] == expected_contents["description"]
    assert response["taskGraph"] == expected_contents["task_graph"]

    assert helpers.is_iso_format(response["createdOn"])
    assert helpers.is_iso_format(response["lastModifiedOn"])

    # Validate PluginRef structure
    for plugin in response["plugins"]:
        assert isinstance(plugin["id"], int)
        assert isinstance(plugin["snapshotId"], int)
        assert isinstance(plugin["name"], str)
        assert isinstance(plugin["url"], str)
        for plugin_file in plugin["files"]:
            assert isinstance(plugin_file["id"], int)
            assert isinstance(plugin_file["snapshotId"], int)
            assert isinstance(plugin_file["filename"], str)
            assert isinstance(plugin_file["url"], str)

    # Validate the Queue IDs structure
    for queue in response["queues"]:
        assert isinstance(queue["id"], int)
        assert isinstance(queue["group"], dict)
        assert isinstance(queue["url"], str)

    # Validate the UserRef structure
    assert isinstance(response["user"]["id"], int)
    assert isinstance(response["user"]["username"], str)
    assert isinstance(response["user"]["url"], str)
    assert response["user"]["id"] == expected_contents["user_id"]

    # Validate the GroupRef structure
    assert isinstance(response["group"]["id"], int)
    assert isinstance(response["group"]["name"], str)
    assert isinstance(response["group"]["url"], str)
    assert response["group"]["id"] == expected_contents["group_id"]

    # Validate the TagRef structure
    for tag in response["tags"]:
        assert isinstance(tag["id"], int)
        assert isinstance(tag["name"], str)
        assert isinstance(tag["url"], str)
    assert response["tags"] == expected_contents["tags"]

    # Validate the EntryPointParameter structure
    for param in response["parameters"]:
        assert isinstance(param["name"], str)
        assert isinstance(param["defaultValue"], str)
        assert isinstance(param["parameterType"], str)
    assert response["parameters"] == expected_contents["parameters"]


def assert_retrieving_entrypoint_by_id_works(
    client: FlaskClient,
    entrypoint_id: int,
    expected: dict[str, Any],
) -> None:
    """Assert that retrieving a entrypoint by id works.

    Args:
        client: The Flask test client.
        entrypoint_id: The id of the entrypoint to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_ENTRYPOINTS_ROUTE}/{entrypoint_id}", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_retrieving_entrypoints_works(
    client: FlaskClient,
    expected: list[dict[str, Any]],
    group_id: int | None = None,
    search: str | None = None,
    paging_info: dict[str, Any] | None = None,
) -> None:
    """Assert that retrieving all entrypoints works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.
        group_id: The group ID used in query parameters.
        search: The search string used in query parameters.
        paging_info: The paging information used in query parameters.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    if group_id is not None:
        query_string["groupId"] = group_id

    if search is not None:
        query_string["search"] = search

    if paging_info is not None:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{V1_ENTRYPOINTS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )

    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_sorting_entrypoint_works(
    client: FlaskClient,
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Assert that entrypoints can be sorted by column ascending/descending.

    Args:
        client: The Flask test client.
        expected: The expected order of entrypoints ids after sorting.
            See test_entrypoint_sort for expected orders.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    query_string["sortBy"] = sortBy
    query_string["descending"] = descending

    response = client.get(
        f"/{V1_ROOT}/{V1_ENTRYPOINTS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )

    response_data = response.get_json()
    entrypoint_ids = [entrypoint["id"] for entrypoint in response_data["data"]]

    assert response.status_code == 200 and entrypoint_ids == expected


def assert_registering_existing_entrypoint_name_fails(
    client: FlaskClient,
    name: str,
    description: str,
    group_id: int,
    task_graph: str,
    parameters: list[dict[str, Any]],
    plugin_ids: list[int],
    queue_ids: list[int],
) -> None:
    """Assert that registering a entrypoint with an existing name fails.

    Args:
        client: The Flask test client.
        name: The name to assign to the new entrypoint.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = actions.register_entrypoint(
        client,
        name=name,
        description=description,
        group_id=group_id,
        task_graph=task_graph,
        parameters=parameters,
        plugin_ids=plugin_ids,
        queue_ids=queue_ids,
    )
    assert response.status_code == 409


def assert_entrypoint_name_matches_expected_name(
    client: FlaskClient, entrypoint_id: int, expected_name: str
) -> None:
    """Assert that the name of a entrypoint matches the expected name.

    Args:
        client: The Flask test client.
        entrypoint_id: The id of the entrypoint to retrieve.
        expected_name: The expected name of the entrypoint.

    Raises:
        AssertionError: If the response status code is not 200 or if the name of the
            entrypoint does not match the expected name.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_ENTRYPOINTS_ROUTE}/{entrypoint_id}",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["name"] == expected_name


def assert_entrypoint_is_not_found(
    client: FlaskClient,
    entrypoint_id: int,
) -> None:
    """Assert that a entrypoint is not found.

    Args:
        client: The Flask test client.
        entrypoint_id: The id of the entrypoint to retrieve.

    Raises:
        AssertionError: If the response status code is not 404.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_ENTRYPOINTS_ROUTE}/{entrypoint_id}",
        follow_redirects=True,
    )
    assert response.status_code == 404


def assert_entrypoint_is_not_associated_with_experiment(
    client: FlaskClient,
    experiment_id: int,
    entrypoint_id: int,
) -> None:
    """Assert that an entrypoint is not associated with an experiment

    Args:
        client: The Flask test client.
        experiment_id: The id of the experiment to retrieve.
        entrypoint_id: The id of the entrypoint to check for association.

    Raises:
        AssertionError: If the response status code is not 200 or if the queue id
            is in the list of queues associated with the entrypoint.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_EXPERIMENTS_ROUTE}/{experiment_id}",
        follow_redirects=True,
    )
    experiment = response.get_json()
    print(experiment)
    entrypoint_ids = set(entrypoint["id"] for entrypoint in experiment["entrypoints"])

    assert response.status_code == 200 and entrypoint_id not in entrypoint_ids


def assert_cannot_rename_entrypoint_with_existing_name(
    client: FlaskClient,
    entrypoint_id: int,
    existing_name: str,
    existing_description: str,
    existing_task_graph: str,
    existing_parameters: list[dict[str, Any]],
    existing_queue_ids: list[int],
) -> None:
    """Assert that renaming a entrypoint with an existing name fails.

    Args:
        client: The Flask test client.
        entrypoint_id: The id of the entrypoint to rename.
        existing_name: str,
        existing_description: str,
        existing_task_graph: str,
        existing_parameters: list[dict[str, Any]],
        existing_queue_ids: list[int],

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = modify_entrypoint(
        client,
        entrypoint_id=entrypoint_id,
        new_name=existing_name,
        new_description=existing_description,
        new_task_graph=existing_task_graph,
        new_parameters=existing_parameters,
        new_queue_ids=existing_queue_ids,
    )
    assert response.status_code == 400


def assert_entrypoint_must_have_unique_param_names(
    client: FlaskClient,
    name: str,
    description: str,
    group_id: int,
    task_graph: str,
    parameters: list[dict[str, Any]],
    plugin_ids: list[int],
    queue_ids: list[int],
) -> None:
    response = actions.register_entrypoint(
        client,
        name=name,
        description=description,
        group_id=group_id,
        task_graph=task_graph,
        parameters=parameters,
        plugin_ids=plugin_ids,
        queue_ids=queue_ids,
    )
    assert response.status_code == 409


# -- Tests -----------------------------------------------------------------------------


def test_create_entrypoint(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_with_files: dict[str, Any],
    registered_queues: dict[str, Any],
) -> None:
    """Test that entrypoints can be correctly registered and retrieved using the API.

    Given an authenticated user, registered plugins, and registered queues, this test
    validates the following sequence of actions:

    - The user registers a entrypoint named "tensorflow_cpu".
    - The response is valid matches the expected values given the registration request.
    - The user is able to retrieve information about the entrypoint using the entrypoint
      id.
    """
    name = "my_entrypoint"
    description = "The first entrypoint."
    user_id = auth_account["id"]
    group_id = auth_account["groups"][0]["id"]
    task_graph = textwrap.dedent(
        """# my entrypoint graph
        graph:
          message:
            my_entrypoint: $name
        """
    )
    parameters = [
        {
            "name": "my_entrypoint_param_1",
            "defaultValue": "my_value",
            "parameterType": "string",
        },
        {
            "name": "my_entrypoint_param_2",
            "defaultValue": "my_value",
            "parameterType": "string",
        },
    ]
    plugin_ids = [registered_plugin_with_files["plugin"]["id"]]
    queue_ids = [queue["id"] for queue in list(registered_queues.values())]
    entrypoint_response = actions.register_entrypoint(
        client,
        name=name,
        description=description,
        group_id=group_id,
        task_graph=task_graph,
        parameters=parameters,
        plugin_ids=plugin_ids,
        queue_ids=queue_ids,
    )
    entrypoint_expected = entrypoint_response.get_json()
    assert_entrypoint_response_contents_matches_expectations(
        response=entrypoint_expected,
        expected_contents={
            "name": name,
            "description": description,
            "user_id": user_id,
            "group_id": group_id,
            "task_graph": task_graph,
            "parameters": parameters,
            "plugin_ids": plugin_ids,
            "queue_ids": queue_ids,
            "tags": [],
        },
    )
    assert_retrieving_entrypoint_by_id_works(
        client, entrypoint_id=entrypoint_expected["id"], expected=entrypoint_expected
    )

    # Testing that parameter names must be unique
    bad_parameters = [
        {
            "name": "not_unique",
            "defaultValue": "my_value",
            "parameterType": "string",
        },
        {
            "name": "not_unique",
            "defaultValue": "my_value",
            "parameterType": "string",
        },
    ]
    assert_entrypoint_must_have_unique_param_names(
        client,
        name=name,
        description=description,
        group_id=group_id,
        task_graph=task_graph,
        parameters=bad_parameters,
        plugin_ids=plugin_ids,
        queue_ids=queue_ids,
    )


def test_entrypoint_get_all(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_entrypoints: dict[str, Any],
) -> None:
    """Test that all entrypoints can be retrieved.

    Given an authenticated user and registered entrypoints, this test validates the
    following sequence of actions:

    - A user registers three entrypoints:
      "tensorflow_cpu",
      "tensorflow_gpu",
      "pytorch_cpu".
    - The user is able to retrieve a list of all registered entrypoints.
    - The returned list of entrypoints matches the full list of registered entrypoints.
    """
    entrypoint_expected_list = list(registered_entrypoints.values())[:3]
    assert_retrieving_entrypoints_works(client, expected=entrypoint_expected_list)


@pytest.mark.parametrize(
    "sortBy, descending , expected",
    [
        (None, None, ["entrypoint1", "entrypoint2", "entrypoint3"]),
        ("name", True, ["entrypoint2", "entrypoint3", "entrypoint1"]),
        ("name", False, ["entrypoint1", "entrypoint3", "entrypoint2"]),
        ("createdOn", True, ["entrypoint3", "entrypoint2", "entrypoint1"]),
        ("createdOn", False, ["entrypoint1", "entrypoint2", "entrypoint3"]),
    ],
)
def test_entrypoint_sort(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_entrypoints: dict[str, Any],
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Test that entrypoints can be sorted by column.

    Given an authenticated user and registered entrypoints, this test validates the
    following sequence of actions:

    - A user registers three entrypoints:
      "entrypoint_one",
      "entrypoint_two",
      "entrypoint_three".
    - The user is able to retrieve a list of all registered entrypoints sorted by a
      column ascending/descending.
    - The returned list of entrypoints matches the order in the parametrize lists above.
    """

    expected_ids = [
        registered_entrypoints[expected_name]["id"] for expected_name in expected
    ]
    assert_sorting_entrypoint_works(client, sortBy, descending, expected=expected_ids)


def test_entrypoint_search_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_entrypoints: dict[str, Any],
) -> None:
    """Test that entrypoints can be queried with a search term.

    Given an authenticated user and registered entrypoints, this test validates the
    following sequence of actions:

    - The user is able to retrieve a list of all registered entrypoints with various
      queries.
    - The returned list of entrypoints matches the expected matches from the query.
    """
    entrypoint_expected_list = list(registered_entrypoints.values())[:2]
    assert_retrieving_entrypoints_works(
        client, expected=entrypoint_expected_list, search="description:*entrypoint*"
    )
    entrypoint_expected_list = list(registered_entrypoints.values())[:3]
    assert_retrieving_entrypoints_works(
        client, expected=entrypoint_expected_list, search="*"
    )


def test_entrypoint_group_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_entrypoints: dict[str, Any],
) -> None:
    """Test that entrypoints can retrieved using a group filter.

    Given an authenticated user and registered entrypoints, this test validates the
    following sequence of actions:

    - The user is able to retrieve a list of all registered entrypoints that are owned
      by the default group.
    - The returned list of entrypoints matches the expected list owned by the default
      group.
    """
    entrypoint_expected_list = list(registered_entrypoints.values())[:3]
    assert_retrieving_entrypoints_works(
        client,
        expected=entrypoint_expected_list,
        group_id=auth_account["groups"][0]["id"],
    )


def test_cannot_register_existing_entrypoint_name(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_entrypoints: dict[str, Any],
    registered_plugin_with_files: dict[str, Any],
    registered_queues: dict[str, Any],
) -> None:
    """Test that registering a entrypoint with an existing name fails.

    Given an authenticated user and registered entrypoints, this test validates the
    following sequence of actions:

    - The user attempts to register a second entrypoint with the same name.
    - The request fails with an appropriate error message and response code.
    """
    existing_entrypoint = registered_entrypoints["entrypoint1"]
    plugin_ids = [registered_plugin_with_files["plugin"]["id"]]
    queue_ids = [queue["id"] for queue in list(registered_queues.values())]

    assert_registering_existing_entrypoint_name_fails(
        client,
        name=existing_entrypoint["name"],
        description="",
        group_id=existing_entrypoint["group"]["id"],
        task_graph="",
        parameters=[],
        plugin_ids=plugin_ids,
        queue_ids=queue_ids,
    )


def test_rename_entrypoint(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_entrypoints: dict[str, Any],
) -> None:
    """Test that a entrypoint can be renamed.

    Given an authenticated user and registered entrypoints, this test validates the
    following sequence of actions:

    - The user issues a request to change the name of a entrypoint.
    - The user retrieves information about the same entrypoint and it reflects the name
      change.
    - The user issues a request to change the name of the entrypoint to the existing
      name.
    - The user retrieves information about the same entrypoint and verifies the name
      remains unchanged.
    - The user issues a request to change the name of a entrypoint to an existing
      entrypoint's name.
    - The request fails with an appropriate error message and response code.
    """
    updated_entrypoint_name = "new_entrypoint_name"
    entrypoint_to_rename = registered_entrypoints["entrypoint1"]
    existing_entrypoint = registered_entrypoints["entrypoint2"]
    queue_ids = [queue["id"] for queue in entrypoint_to_rename["queues"]]

    modified_entrypoint = modify_entrypoint(
        client,
        entrypoint_id=entrypoint_to_rename["id"],
        new_name=updated_entrypoint_name,
        new_description=entrypoint_to_rename["description"],
        new_task_graph=entrypoint_to_rename["taskGraph"],
        new_parameters=entrypoint_to_rename["parameters"],
        new_queue_ids=queue_ids,
    ).get_json()
    assert_entrypoint_name_matches_expected_name(
        client,
        entrypoint_id=entrypoint_to_rename["id"],
        expected_name=updated_entrypoint_name,
    )
    entrypoint_expected_list = [
        modified_entrypoint,
        registered_entrypoints["entrypoint2"],
        registered_entrypoints["entrypoint3"],
    ]
    assert_retrieving_entrypoints_works(client, expected=entrypoint_expected_list)

    modified_entrypoint = modify_entrypoint(
        client,
        entrypoint_id=entrypoint_to_rename["id"],
        new_name=updated_entrypoint_name,
        new_description=entrypoint_to_rename["description"],
        new_task_graph=entrypoint_to_rename["taskGraph"],
        new_parameters=entrypoint_to_rename["parameters"],
        new_queue_ids=entrypoint_to_rename["queues"],
    ).get_json()
    assert_entrypoint_name_matches_expected_name(
        client,
        entrypoint_id=entrypoint_to_rename["id"],
        expected_name=updated_entrypoint_name,
    )

    assert_cannot_rename_entrypoint_with_existing_name(
        client,
        entrypoint_id=entrypoint_to_rename["id"],
        existing_name=existing_entrypoint["name"],
        existing_description=entrypoint_to_rename["description"],
        existing_task_graph=entrypoint_to_rename["taskGraph"],
        existing_parameters=entrypoint_to_rename["parameters"],
        existing_queue_ids=entrypoint_to_rename["queues"],
    )


def test_delete_entrypoint_by_id(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_experiments: dict[str, Any],
) -> None:
    """Test that a entrypoint can be deleted by referencing its id.

    Given an authenticated user and registered entrypoints, this test validates the
    following sequence of actions:

    - The user deletes a entrypoint by referencing its id.
    - The user attempts to retrieve information about the deleted entrypoint.
    - The request fails with an appropriate error message and response code.
    - The entrypoint is no longer associated with the experiment.
    """
    experiment = registered_experiments["experiment1"]
    entrypoint_to_delete = experiment["entrypoints"][0]

    delete_entrypoint(client, entrypoint_id=entrypoint_to_delete["id"])
    assert_entrypoint_is_not_found(client, entrypoint_id=entrypoint_to_delete["id"])
    assert_entrypoint_is_not_associated_with_experiment(
        client, experiment_id=experiment["id"], entrypoint_id=entrypoint_to_delete["id"]
    )


def test_manage_existing_entrypoint_draft(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_entrypoints: dict[str, Any],
) -> None:
    """Test that a draft of an existing entrypoint can be created and managed by
        the user

    Given an authenticated user and registered entrypoints, this test validates the
    following sequence of actions:

    - The user creates a draft of an existing entrypoint
    - The user retrieves information about the draft and gets the expected response
    - The user attempts to create another draft of the same existing entrypoint
    - The request fails with an appropriate error message and response code.
    - The user modifies the name of the entrypoint in the draft
    - The user retrieves information about the draft and gets the expected response
    - The user deletes the draft
    - The user attempts to retrieve information about the deleted draft.
    - The request fails with an appropriate error message and response code.
    """
    entrypoint = registered_entrypoints["entrypoint1"]
    name = "draft"
    new_name = "draft2"
    description = "description"
    task_graph = textwrap.dedent(
        """# my entrypoint graph
        graph:
          message:
            my_entrypoint: $name
        """
    )
    parameters = [
        {
            "name": "my_entrypoint_param",
            "defaultValue": "my_value",
            "parameterType": "string",
        }
    ]
    plugin_ids = [plugin["id"] for plugin in entrypoint["plugins"]]
    queue_ids = [queue["id"] for queue in entrypoint["queues"]]

    # test creation
    payload = {
        "name": name,
        "description": description,
        "taskGraph": task_graph,
        "parameters": parameters,
        "plugins": plugin_ids,
        "queues": queue_ids,
    }
    expected = {
        "user_id": auth_account["id"],
        "group_id": entrypoint["group"]["id"],
        "resource_id": entrypoint["id"],
        "resource_snapshot_id": entrypoint["snapshot"],
        "num_other_drafts": 0,
        "payload": payload,
    }
    response = actions.create_existing_resource_draft(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        resource_id=entrypoint["id"],
        payload=payload,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(response, expected)
    asserts.assert_retrieving_draft_by_resource_id_works(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        resource_id=entrypoint["id"],
        expected=response,
    )
    asserts.assert_creating_another_existing_draft_fails(
        client, resource_route=V1_ENTRYPOINTS_ROUTE, resource_id=entrypoint["id"]
    )

    # test modification
    payload = {"name": new_name, "description": description, "taskGraph": task_graph}
    expected = {
        "user_id": auth_account["id"],
        "group_id": entrypoint["group"]["id"],
        "resource_id": entrypoint["id"],
        "resource_snapshot_id": entrypoint["snapshot"],
        "num_other_drafts": 0,
        "payload": payload,
    }
    response = actions.modify_existing_resource_draft(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        resource_id=entrypoint["id"],
        payload=payload,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(response, expected)

    # test deletion
    actions.delete_existing_resource_draft(
        client, resource_route=V1_ENTRYPOINTS_ROUTE, resource_id=entrypoint["id"]
    )
    asserts.assert_existing_draft_is_not_found(
        client, resource_route=V1_ENTRYPOINTS_ROUTE, resource_id=entrypoint["id"]
    )


def test_manage_new_entrypoint_drafts(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> None:
    """Test that drafts of entrypoint can be created and managed by the user

    Given an authenticated user, this test validates the following sequence of actions:

    - The user creates two entrypoint drafts
    - The user retrieves information about the drafts and gets the expected response
    - The user modifies the description of the entrypoint in the first draft
    - The user retrieves information about the draft and gets the expected response
    - The user deletes the first draft
    - The user attempts to retrieve information about the deleted draft.
    - The request fails with an appropriate error message and response code.
    """
    group_id = auth_account["groups"][0]["id"]
    drafts = {
        "draft1": {
            "name": "entrypoint1",
            "description": "new entrypoint",
            "taskGraph": "graph",
        },
        "draft2": {
            "name": "entrypoint2",
            "description": "entrypoint",
            "taskGraph": "graph",
            "queues": [1, 3],
            "plugins": [2],
        },
    }

    # test creation
    draft1_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": drafts["draft1"],
    }
    draft1_response = actions.create_new_resource_draft(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        group_id=group_id,
        payload=drafts["draft1"],
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        draft1_response, draft1_expected
    )
    asserts.assert_retrieving_draft_by_id_works(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        draft_id=draft1_response["id"],
        expected=draft1_response,
    )
    draft2_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": drafts["draft2"],
    }
    draft2_response = actions.create_new_resource_draft(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        group_id=group_id,
        payload=drafts["draft2"],
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        draft2_response, draft2_expected
    )
    asserts.assert_retrieving_draft_by_id_works(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        draft_id=draft2_response["id"],
        expected=draft2_response,
    )
    asserts.assert_retrieving_drafts_works(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        expected=[draft1_response, draft2_response],
    )

    # test modification
    draft1_mod = {
        "name": "draft1",
        "description": "new description",
        "taskGraph": "graph",
    }
    draft1_mod_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": draft1_mod,
    }
    response = actions.modify_new_resource_draft(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        draft_id=draft1_response["id"],
        payload=draft1_mod,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        response, draft1_mod_expected
    )

    # test deletion
    actions.delete_new_resource_draft(
        client, resource_route=V1_ENTRYPOINTS_ROUTE, draft_id=draft1_response["id"]
    )
    asserts.assert_new_draft_is_not_found(
        client, resource_route=V1_ENTRYPOINTS_ROUTE, draft_id=draft1_response["id"]
    )


def test_manage_entrypoint_snapshots(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_entrypoints: dict[str, Any],
) -> None:
    """Test that different snapshots of a entrypoint can be retrieved by the user.

    Given an authenticated user and registered entrypoints, this test validates the
    following sequence of actions:

    - The user modifies a entrypoint
    - The user retrieves information about the original snapshot of the entrypoint and
      gets the expected response
    - The user retrieves information about the new snapshot of the entrypoint and gets
      the expected response
    - The user retrieves a list of all snapshots of the entrypoint and gets the expected
      response
    """
    entrypoint_to_rename = registered_entrypoints["entrypoint1"]
    queue_ids = [queue["id"] for queue in entrypoint_to_rename["queues"]]
    modified_entrypoint = modify_entrypoint(
        client,
        entrypoint_id=entrypoint_to_rename["id"],
        new_name=entrypoint_to_rename["name"] + "modified",
        new_description=entrypoint_to_rename["description"],
        new_task_graph=entrypoint_to_rename["taskGraph"],
        new_parameters=entrypoint_to_rename["parameters"],
        new_queue_ids=queue_ids,
    ).get_json()
    entrypoint_to_rename["latestSnapshot"] = False
    entrypoint_to_rename["lastModifiedOn"] = modified_entrypoint["lastModifiedOn"]
    entrypoint_to_rename.pop("queues")
    entrypoint_to_rename.pop("hasDraft")
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        resource_id=entrypoint_to_rename["id"],
        snapshot_id=entrypoint_to_rename["snapshot"],
        expected=entrypoint_to_rename,
    )
    modified_entrypoint.pop("queues")
    modified_entrypoint.pop("hasDraft")
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        resource_id=modified_entrypoint["id"],
        snapshot_id=modified_entrypoint["snapshot"],
        expected=modified_entrypoint,
    )
    expected_snapshots = [entrypoint_to_rename, modified_entrypoint]
    asserts.assert_retrieving_snapshots_works(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        resource_id=entrypoint_to_rename["id"],
        expected=expected_snapshots,
    )


def test_tag_entrypoint(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_entrypoints: dict[str, Any],
    registered_tags: dict[str, Any],
) -> None:
    """Test that different versions of a entrypoint can be retrieved by the user.

    Given an authenticated user and registered entrypoints, this test validates the
    following sequence of actions:

    """
    entrypoint = registered_entrypoints["entrypoint1"]
    tags = [tag["id"] for tag in registered_tags.values()]

    # test append
    response = actions.append_tags(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        resource_id=entrypoint["id"],
        tag_ids=[tags[0], tags[1]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1]]
    )
    response = actions.append_tags(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        resource_id=entrypoint["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1], tags[2]]
    )

    # test remove
    actions.remove_tag(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        resource_id=entrypoint["id"],
        tag_id=tags[1],
    )
    response = actions.get_tags(
        client, resource_route=V1_ENTRYPOINTS_ROUTE, resource_id=entrypoint["id"]
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[2]]
    )

    # test modify
    response = actions.modify_tags(
        client,
        resource_route=V1_ENTRYPOINTS_ROUTE,
        resource_id=entrypoint["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[1], tags[2]]
    )

    # test delete
    response = actions.remove_tags(
        client, resource_route=V1_ENTRYPOINTS_ROUTE, resource_id=entrypoint["id"]
    )
    response = actions.get_tags(
        client, resource_route=V1_ENTRYPOINTS_ROUTE, resource_id=entrypoint["id"]
    )
    asserts.assert_tags_response_contents_matches_expectations(response.get_json(), [])



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The schemas for serializing/deserializing Job resources."""
from marshmallow import Schema, fields, validate

from dioptra.restapi.v1.artifacts.schema import ArtifactRefSchema
from dioptra.restapi.v1.schemas import (
    BasePageSchema,
    GroupIdQueryParametersSchema,
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
    generate_base_resource_ref_schema,
    generate_base_resource_schema,
)

JobRefSchema = generate_base_resource_ref_schema("Job")
JobSnapshotRefSchema = generate_base_resource_ref_schema("Job", keep_snapshot_id=True)


class JobMlflowRunSchema(Schema):
    """The schema for the data in a mlflowRun resource."""

    mlflowRunId = fields.UUID(
        attribute="mlflow_run_id",
        metadata=dict(description="UUID for the associated Mlflow Run."),
    )


class JobStatusSchema(Schema):
    """The fields schema for the data in a Job status resource."""

    id = fields.Integer(
        attribute="id",
        metadata=dict(description="ID for the Job resource."),
        dump_only=True,
    )
    status = fields.String(
        attribute="status",
        validate=validate.OneOf(
            ["queued", "started", "deferred", "finished", "failed"],
        ),
        metadata=dict(
            description="The current status of the job. The allowed values are: "
            "queued, started, deferred, finished, failed.",
        ),
    )


JobBaseSchema = generate_base_resource_schema("Job", snapshot=True)


class JobSchema(JobBaseSchema):  # type: ignore
    """The schema for the data stored in a Job resource."""

    from dioptra.restapi.v1.entrypoints.schema import EntrypointSnapshotRefSchema
    from dioptra.restapi.v1.experiments.schema import ExperimentSnapshotRefSchema
    from dioptra.restapi.v1.queues.schema import QueueSnapshotRefSchema

    description = fields.String(
        attribute="description",
        metadata=dict(description="Description of the Job resource."),
        load_default=None,
    )
    queueId = fields.Integer(
        attribute="queue_id",
        data_key="queue",
        metadata=dict(description="An integer identifying a registered queue."),
        load_only=True,
        required=True,
    )
    queue = fields.Nested(
        QueueSnapshotRefSchema,
        attribute="queue",
        metadata=dict(description="The active queue used to run the Job."),
        dump_only=True,
    )
    experiment = fields.Nested(
        ExperimentSnapshotRefSchema,
        attribute="experiment",
        metadata=dict(description="The registered experiment associated with the Job."),
        dump_only=True,
    )
    entrypointId = fields.Integer(
        attribute="entrypoint_id",
        data_key="entrypoint",
        metadata=dict(description="An integer identifying a registered entry point."),
        load_only=True,
        required=True,
    )
    entrypoint = fields.Nested(
        EntrypointSnapshotRefSchema,
        attribute="entrypoint",
        metadata=dict(description="The entry point associated with the Job."),
        dump_only=True,
    )
    values = fields.Dict(
        keys=fields.String(),
        values=fields.String(),
        attribute="values",
        allow_none=True,
        metadata=dict(
            description=(
                "A dictionary of keyword arguments to pass to the Job's Entrypoint."
            ),
        ),
    )
    timeout = fields.String(
        attribute="timeout",
        load_default="24h",
        metadata=dict(
            description="The maximum alloted time for a job before it times out and "
            "is stopped. If omitted, the job timeout will default to 24 hours.",
        ),
    )
    status = fields.String(
        attribute="status",
        metadata=dict(
            description="The current status of the job. The allowed values are: "
            "queued, started, deferred, finished, failed.",
        ),
        dump_only=True,
    )
    artifacts = fields.Nested(
        ArtifactRefSchema,
        attribute="artifacts",
        many=True,
        metadata=dict(description="Artifacts created by the Job resource."),
        dump_only=True,
    )


class JobPageSchema(BasePageSchema):
    """The paged schema for the data stored in a Job resource."""

    data = fields.Nested(
        JobSchema,
        many=True,
        metadata=dict(description="List of Job resources in the current page."),
    )


class JobGetQueryParameters(
    PagingQueryParametersSchema,
    GroupIdQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
):
    """The query parameters for the GET method of the /jobs endpoint."""


class ExperimentJobGetQueryParameters(
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
):
    """The query parameters for the GET method of the /experiments/{id}/jobs
    endpoint."""



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import graphlib
from collections.abc import Callable, Container, Iterator, Mapping, Sequence
from typing import Any, Optional, Union

import jsonschema.validators

from dioptra.sdk.exceptions.task_engine import (
    StepNotFoundError,
    StepReferenceCycleError,
)
from dioptra.task_engine.error_message import validation_error_to_message


def is_iterable(value: Any) -> bool:
    """
    Determine whether the given value is iterable.  Works by attempting to
    get an iterator from it.

    Args:
        value: The value to check

    Returns:
        True if the value is iterable; False if not
    """
    try:
        iter(value)
    except TypeError:
        result = False
    else:
        result = True

    return result


def schema_validate(
    instance: Any,
    schema: Union[dict[str, Any], bool],
    location_desc_callback: Optional[Callable[[Sequence[Union[int, str]]], str]] = None,
) -> list[str]:
    """
    Validate the given instance against the given JSON-Schema.

    Args:
        instance: A value to validate
        schema: JSON-Schema as a data structure, e.g. parsed JSON
        location_desc_callback: A callback function used to customize the
            description of the location of errors.  Takes a programmatic "path"
            structure as a sequence of strings/ints, and should return a nice
            one-line string description.  Defaults to a simple generic
            implementation which produces descriptions which aren't very nice.

    Returns:
        A list of error messages; will be empty if validation succeeded
    """
    # Make use of a more complex API to try to produce better schema
    # validation error messages.
    validator_class = jsonschema.validators.validator_for(schema)
    validator = validator_class(schema=schema)

    error_messages = [
        validation_error_to_message(error, schema, location_desc_callback)
        for error in validator.iter_errors(instance)
    ]

    return error_messages


def step_get_plugin_short_name(step: Mapping[str, Any]) -> Optional[str]:
    """
    Get the plugin short name from a step description.  There is a bit of
    complexity to this, since step descriptions can take different forms, and
    some properties can have special meanings and should not be mistaken for a
    task plugin short name.

    Args:
        step: A step description, as a mapping

    Returns:
        A task plugin short name, or None if one was not found
    """
    plugin_name = None
    if "task" in step:
        plugin_name = step["task"]
    else:
        for key in step:
            # Top-level "dependencies" key has a special meaning; ignore that.
            if key != "dependencies":
                plugin_name = key
                break

    return plugin_name


def step_get_invocation_arg_specs(
    step_def: Mapping[str, Any]
) -> Union[tuple[list[Any], Mapping[str, Any]], tuple[None, None]]:
    """
    Get invocation positional and keyword arg specs from the given step.  This
    just extracts and returns the specs as they are and doesn't resolve any
    references.

    :param step_def: A task graph step definition
    :return: A 2-tuple consisting of positional args and keyword args.
        Positional args are returned as a list of values; keyword args are
        returned as a mapping from name to value.  Either collection may be
        empty.  If no arg specs could be found in the step definition, (None,
        None) is returned.  This is different from empty arg specs.  If no arg
        specs are found, that means the step definition is malformed.
    """

    task_plugin_short_name = step_get_plugin_short_name(step_def)
    if task_plugin_short_name is None:
        pos_arg_specs = kwarg_specs = None

    else:
        if "task" in step_def:
            # Mixed step definition
            pos_arg_specs = step_def.get("args", [])
            kwarg_specs = step_def.get("kwargs", {})

        else:
            # Positional or keyword step definition
            arg_specs = step_def[task_plugin_short_name]

            if isinstance(arg_specs, dict):
                # Must be keyword
                pos_arg_specs = []
                kwarg_specs = arg_specs

            else:
                # Must be positional
                pos_arg_specs = arg_specs
                kwarg_specs = {}

        if not isinstance(pos_arg_specs, list):
            # ensure positional args are in a list
            pos_arg_specs = [pos_arg_specs]

    return pos_arg_specs, kwarg_specs


def is_reference(value: str) -> bool:
    """
    Determine whether the given string, as may be found in a task invocation
    specification, is syntactically a reference.  References begin with a
    dollar sign and are followed by some non-dollar-sign characters, e.g.
    "$foo".  A dollar sign may be escaped by doubling it.  So "$$foo" is not a
    reference, but just means the string "$foo".  If there are no
    non-dollar-sign characters, it is not a reference, e.g. "$", "$$", etc.
    This function does not try to actually resolve the reference, so if the
    value is a reference, this function does not guarantee it is valid.

    Args:
        value: The string to check

    Returns:
        True if the given string is a reference; False if not
    """

    return value != "$" and value.startswith("$") and not value.startswith("$$")


def get_references(input_: Any) -> Iterator[str]:
    """
    Search for references within a task invocation specification, and generate
    the names.  This just makes the reference determination syntactically, and
    doesn't attempt to resolve any reference.

    Args:
        input_: The arg spec structure to search for references

    Yields:
        References, as strings without the leading "$"
    """

    if isinstance(input_, str):
        if is_reference(input_):
            yield input_[1:]  # drop the leading "$"

    elif isinstance(input_, dict):
        if "task" in input_:
            args = input_.get("args")
            kwargs = input_.get("kwargs")
            if args:
                yield from get_references(args)
            if kwargs:
                yield from get_references(kwargs.values())
        else:
            yield from get_references(input_.values())

    elif is_iterable(input_):
        for elt in input_:
            yield from get_references(elt)


def _get_step_references(input_: Any, step_names: Container[str]) -> Iterator[str]:
    """
    Generate step names from references in the input which refer to steps.

    Args:
        input_: The arg spec structure to search for references to steps
        step_names: A container with all of the step names.  This allows us
            to recognize a step name when we see one.

    Yields:
        Referenced step names
    """
    # Note: erroneous references will not be caught here (i.e. those which
    # don't refer to either a step name or a global parameter).  Those errors
    # will be caught later, when we actually need to resolve the references.
    for ref_name in get_references(input_):
        dot_idx = ref_name.find(".")
        if dot_idx >= 0:
            step_name = ref_name[:dot_idx]
        else:
            step_name = ref_name

        if step_name in step_names:
            yield step_name


def get_sorted_steps(step_graph: Mapping[str, Any]) -> list[str]:
    """
    Find a topological sorted list of step names for the given graph.

    Args:
        step_graph: Step definitions, as a mapping from step name to step
            definition.

    Returns:
        A list of step names
    """
    topo_sorter: graphlib.TopologicalSorter = graphlib.TopologicalSorter()

    for step_name, step_def in step_graph.items():
        topo_sorter.add(step_name)

        for dep_step_name in _get_step_references(step_def, step_graph.keys()):
            if dep_step_name in step_graph:
                topo_sorter.add(step_name, dep_step_name)
            else:
                raise StepNotFoundError(dep_step_name, step_name)

        explicit_deps = step_def.get("dependencies", [])
        if isinstance(explicit_deps, str):
            explicit_deps = [explicit_deps]

        for dep_step_name in explicit_deps:
            if dep_step_name in step_graph:
                topo_sorter.add(step_name, dep_step_name)
            else:
                raise StepNotFoundError(dep_step_name, step_name)

    try:
        sorted_steps = list(topo_sorter.static_order())
    except graphlib.CycleError as e:
        raise StepReferenceCycleError(e.args[1]) from e

    return sorted_steps


def input_def_get_name_type(in_def: Mapping[str, Any]) -> tuple[str, str]:
    """
    Get a parameter name and type from a task input parameter definition.

    :param in_def: A task input definition
    :return: A (name, type) 2-tuple
    """
    if "name" in in_def:
        in_name = in_def["name"]
        in_type = in_def["type"]

    else:
        in_name, in_type = next(iter(in_def.items()))

    return in_name, in_type


def make_task_input_map(task_def: Mapping[str, Any]) -> Mapping[str, Any]:
    """
    Make a mapping from task input parameter name to parameter definition,
    from the given task definition.  Since python dicts remember insertion
    order (as iteration order), if we are deliberate about insertion order, we
    can use the mapping for both positional and keyword arg validation: we can
    be sure that the Nth positional task invocation parameter can be matched to
    the Nth map item in iteration order.

    :param task_def: A task definition
    :return: A mapping from task input parameter name to definition
    """
    task_inputs = task_def.get("inputs", [])

    if not isinstance(task_inputs, list):
        task_inputs = [task_inputs]

    task_inputs_map = {}
    for task_input in task_inputs:
        input_name, _ = input_def_get_name_type(task_input)

        # Assume all these names are unique.
        task_inputs_map[input_name] = task_input

    return task_inputs_map



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import datetime
import uuid
from typing import TYPE_CHECKING

from sqlalchemy import select
from sqlalchemy.orm import Mapped, column_property, mapped_column, relationship

from dioptra.restapi.db.db import (
    datetimetz,
    db,
    guid,
    intpk,
    optionaldatetimetz,
    optionalstr,
    text_,
)

from .constants import user_lock_types
from .locks import UserLock

if TYPE_CHECKING:
    from .groups import Group, GroupManager, GroupMember
    from .resources import DraftResource, ResourceSnapshot, SharedResource
    from .tags import Tag

# -- ORM Classes -----------------------------------------------------------------------


class User(db.Model):  # type: ignore[name-defined]
    __tablename__ = "users"

    # Database fields
    user_id: Mapped[intpk] = mapped_column(init=False)
    alternative_id: Mapped[guid] = mapped_column(
        init=False, nullable=False, unique=True, index=True
    )
    username: Mapped[text_] = mapped_column(nullable=False, index=True)
    password: Mapped[optionalstr]
    email_address: Mapped[text_] = mapped_column(nullable=False, index=True)
    created_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)
    last_modified_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)
    last_login_on: Mapped[optionaldatetimetz] = mapped_column(init=False, nullable=True)
    password_expire_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)

    # Derived fields (read-only)
    is_deleted: Mapped[bool] = column_property(
        select(UserLock.user_id)
        .where(
            UserLock.user_id == user_id,
            UserLock.user_lock_type == user_lock_types.DELETE,
        )
        .correlate_except(UserLock)
        .exists()
    )

    # Relationships
    group_memberships: Mapped[list["GroupMember"]] = relationship(
        init=False, back_populates="user"
    )
    group_managerships: Mapped[list["GroupManager"]] = relationship(
        init=False, back_populates="user"
    )
    created_drafts: Mapped[list["DraftResource"]] = relationship(
        init=False, viewonly=True
    )
    created_groups: Mapped[list["Group"]] = relationship(init=False, viewonly=True)
    created_shares: Mapped[list["SharedResource"]] = relationship(
        init=False, viewonly=True
    )
    created_snapshots: Mapped[list["ResourceSnapshot"]] = relationship(
        init=False, viewonly=True
    )
    created_tags: Mapped[list["Tag"]] = relationship(init=False, viewonly=True)
    locks: Mapped[list["UserLock"]] = relationship(init=False, back_populates="user")

    # Initialize default values using dataclass __post_init__ method
    # https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__
    def __post_init__(self) -> None:
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self.alternative_id = uuid.uuid4()
        self.created_on = timestamp
        self.last_modified_on = timestamp
        self.last_login_on = None
        self.password_expire_on = timestamp + datetime.timedelta(days=365)

    @property
    def is_authenticated(self) -> bool:
        """Return True if the user is authenticated, False otherwise."""
        return self.is_active

    @property
    def is_active(self) -> bool:
        """Return True if the user account is active, False otherwise."""
        return not self.is_deleted

    @property
    def is_anonymous(self) -> bool:
        """Return True if the user is registered, False otherwise."""
        return False

    def get_id(self) -> str:
        """Get the user's session identifier.

        Returns:
            The user's identifier as a string.
        """
        return str(self.alternative_id)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Test suite for user operations.
This module contains a set of tests that validate the CRUD operations and additional
functionalities for the user entity. The tests ensure that the users can be
registered, modified, and deleted as expected through the REST API.
"""
from typing import Any

from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy
from werkzeug.test import TestResponse

from dioptra.restapi.routes import V1_ROOT, V1_USERS_ROUTE

from ..lib import actions, helpers

# -- Actions ---------------------------------------------------------------------------


def modify_current_user(
    client: FlaskClient,
    new_username: str,
    new_email: str,
) -> TestResponse:
    """Change the current user's email using the API.

    Args:
        client The Flask test client.
        new_email: The new email to assign to the user.

    Returns:
        The response from the API.
    """

    payload = {"username": new_username, "email": new_email}

    return client.put(
        f"/{V1_ROOT}/{V1_USERS_ROUTE}/current", json=payload, follow_redirects=True
    )


def delete_current_user(
    client: FlaskClient,
    password: str,
) -> TestResponse:
    """Delete the current user using the API.

    Args:
        client: The Flask test client.

    Returns:
        The response from the API.
    """
    payload = {"password": password}
    return client.delete(
        f"/{V1_ROOT}/{V1_USERS_ROUTE}/current", json=payload, follow_redirects=True
    )


def change_current_user_password(
    client: FlaskClient,
    old_password: str,
    new_password: str,
):
    """Change the current user password using the API."""
    payload = {
        "oldPassword": old_password,
        "newPassword": new_password,
        "confirmNewPassword": new_password,
    }
    return client.post(
        f"/{V1_ROOT}/{V1_USERS_ROUTE}/current/password",
        json=payload,
        follow_redirects=True,
    )


def change_user_password(
    client: FlaskClient, user_id: int, old_password: str, new_password: str
):
    """Change a user password using its ID using the API."""
    payload = {
        "oldPassword": old_password,
        "newPassword": new_password,
        "confirmNewPassword": new_password,
    }
    return client.post(
        f"/{V1_ROOT}/{V1_USERS_ROUTE}/{user_id}/password",
        json=payload,
        follow_redirects=True,
    )


# -- Assertions ----------------------------------------------------------------


def assert_user_response_contents_matches_expectations(
    response: dict[str, Any],
    expected_contents: dict[str, Any],
    current_user: bool = False,
) -> None:
    """Assert that user response contents is valid.

    Args:
        response: The actual response from the API.
        expected_contents: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response or if the response contents is not
            valid.
    """

    expected_keys = {
        "username",
        "email",
        "id",
    }
    if current_user:
        expected_keys.update(
            {
                "groups",
                "createdOn",
                "lastModifiedOn",
                "lastLoginOn",
                "passwordExpiresOn",
            }
        )
    assert set(response.keys()) == expected_keys

    # Validate non-Ref fields
    assert isinstance(response["username"], str)
    assert isinstance(response["email"], str)
    assert isinstance(response["id"], int)

    assert response["username"] == expected_contents["username"]
    assert response["email"] == expected_contents["email"]

    if current_user:
        assert isinstance(response["createdOn"], str)
        assert isinstance(response["lastModifiedOn"], str)
        assert isinstance(response["lastLoginOn"], (str, type(None)))
        assert isinstance(response["passwordExpiresOn"], str)

        assert helpers.is_iso_format(response["createdOn"])
        assert helpers.is_iso_format(response["lastModifiedOn"])
        if response["lastLoginOn"] is not None:
            assert helpers.is_iso_format(response["lastLoginOn"])
        assert helpers.is_iso_format(response["passwordExpiresOn"])

        # Validate the GroupRef structure
        assert isinstance(response["groups"][0]["id"], int)
        assert isinstance(response["groups"][0]["name"], str)
        assert isinstance(response["groups"][0]["url"], str)


def assert_retrieving_user_by_id_works(
    client: FlaskClient, user_id: int, expected: dict[str, Any]
) -> None:
    """Assert that retrieving a user by id works.

    Args:
        client: The Flask test client.
        user_id: The id of the user to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_USERS_ROUTE}/{user_id}", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_retrieving_current_user_works(
    client: FlaskClient,
    expected: dict[str, Any],
) -> None:
    """Assert that retrieving the current user works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(f"/{V1_ROOT}/{V1_USERS_ROUTE}/current", follow_redirects=True)
    to_ignore = ["lastLoginOn", "lastModifiedOn"]
    response_info_filtered = {
        k: v for k, v in response.get_json().items() if k not in to_ignore
    }
    expected_filtered = {k: v for k, v in expected.items() if k not in to_ignore}
    assert response.status_code == 200 and response_info_filtered == expected_filtered


def assert_retrieving_all_users_works(
    client: FlaskClient,
    expected: list[dict[str, Any]],
) -> None:
    """Assert that retrieving all queues works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_USERS_ROUTE}",
        query_string={},
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_retrieving_users_works(
    client: FlaskClient,
    expected: list[dict[str, Any]],
    search: str | None = None,
    paging_info: dict[str, int] | None = None,
) -> None:
    """Assert that retrieving all users works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.
        search: The search string used in query parameters.
        paging_info: The paging information used in query parameters.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    query_string = {}

    if search is not None:
        query_string["search"] = search

    if paging_info is not None:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{V1_USERS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_registering_existing_username_fails(
    client: FlaskClient,
    existing_username: str,
    non_existing_email: str,
) -> None:
    """Assert that registering a user with an existing username fails.

    Args:
        client: The Flask test client.
        username: The username to assign to the new user.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    password = "supersecurepassword"
    response = actions.register_user(
        client, existing_username, non_existing_email, password
    )
    assert response.status_code == 409


def assert_registering_existing_email_fails(
    client: FlaskClient,
    non_existing_username: str,
    existing_email: str,
) -> None:
    """Assert that registering a user with an existing username fails.

    Args:
        client: The Flask test client.
        username: The username to assign to the new user.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    password = "supersecurepassword"
    response = actions.register_user(
        client, non_existing_username, existing_email, password
    )
    assert response.status_code == 409


def assert_user_username_matches_expected_name(
    client: FlaskClient, user_id: int, expected_name: str
) -> None:
    """Assert that the name of a user matches the expected name.

    Args:
        client: The Flask test client.
        user_id: The id of the user to retrieve.
        expected_name: The expected name of the user.

    Raises:
        AssertionError: If the response status code is not 200 or if the name of the
            user does not match the expected name.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_USERS_ROUTE}/{user_id}",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["name"] == expected_name


def assert_current_user_username_matches_expected_name(
    client: FlaskClient, expected_name: str
) -> None:
    """Assert that the name of the current user matches the expected name.

    Args:
        client: The Flask test client.
        expected_name: The expected name of the user.

    Raises:
        AssertionError: If the response status code is not 200 or if the name of the
            user does not match the expected name.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_USERS_ROUTE}/current",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["name"] == expected_name


def assert_user_is_not_found(
    client: FlaskClient,
    user_id: int,
) -> None:
    """Assert that a user is not found.

    Args:
        client: The Flask test client.
        user_id: The id of the user to retrieve.

    Raises:
        AssertionError: If the response status code is not 404.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_USERS_ROUTE}/{user_id}",
        follow_redirects=True,
    )
    assert response.status_code == 404


def assert_cannot_rename_user_with_existing_username(
    client: FlaskClient,
    existing_username: str,
) -> None:
    """Assert that renaming a user with an existing username fails.

    Args:
        client: The Flask test client.
        existing_username: The username of the existing user.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = modify_current_user(
        client=client,
        new_username=existing_username,
        new_email="new_email",
    )
    assert response.status_code == 400


def assert_cannot_rename_user_with_existing_email(
    client: FlaskClient,
    existing_email: str,
) -> None:
    """Assert that changing a user email with an existing email fails.

    Args:
        client: The Flask test client.
        existing_email: The email of the existing user.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = modify_current_user(
        client=client,
        new_username="new_username",
        new_email=existing_email,
    )
    assert response.status_code == 400


def assert_login_works(
    client: FlaskClient,
    username: str,
    password: str,
):
    """Assert that logging in using a username and password works.

    Args:
        client: The Flask test client.
        username: The username of the user to be logged in.
        password: The password of the user to be logged in.

    Raises:
        AssertionError: If the response status code is not 200.
    """
    assert actions.login(client, username, password).status_code == 200


def assert_user_does_not_exist(
    client: FlaskClient,
    username: str,
    password: str,
):
    """Assert that the user does not exist.

    Args:
        client: The Flask test client.
        username: The username of the user to be logged in.
        password: The password of the user to be logged in.

    Raises:
        AssertionError: If the response status code is not 404.
    """
    assert actions.login(client, username, password).status_code == 404


def assert_login_is_unauthorized(
    client: FlaskClient,
    username: str,
    password: str,
):
    """Assert that logging in using a username and password is unauthorized.

    Args:
        client: The Flask test client.
        username: The username of the user to be logged in.
        password: The password of the user to be logged in.

    Raises:
        AssertionError: If the response status code is not 401.
    """
    assert actions.login(client, username, password).status_code == 401


def assert_new_password_cannot_be_existing(
    client: FlaskClient,
    password: str,
    user_id: str = None,
):
    """Assert that changing a user (current or otherwise) password to the
    existing password fails.

    Args:
        client (FlaskClient): The Flask test client.
        password (str): The existing password of the user.
        user_id (str, optional): The user id. If this is not passed,
        we assume we are the current user. Defaults to None.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    # Means we are the current user.
    if not user_id:
        assert (
            change_current_user_password(client, password, password).status_code == 403
        )
    else:
        assert (
            change_user_password(client, user_id, password, password).status_code == 403
        )


# -- Tests -------------------------------------------------------------


def test_create_user(
    client: FlaskClient,
    db: SQLAlchemy,
) -> None:
    """Test that we can create a user and its response is expected.

    This test validates the following sequence of actions:

    - Register a new user using the username, email, and password.
    - Assert that the response matches what we expect to see, i.e. CurrentUser
    schema information.
    - Login with the user that we just created.
    - Assert that retrieving the current user works.
    - Finally, assert that we can retrieve the current user by ID and it matches
    our expectations.
    """
    username = "user"
    email = "user@example.org"
    password = "supersecurepassword"

    # Posting a user returns CurrentUserSchema.
    user_response = actions.register_user(client, username, email, password).get_json()
    assert_user_response_contents_matches_expectations(
        response=user_response,
        expected_contents={
            "username": username,
            "email": email,
        },
        current_user=True,
    )

    actions.login(client, username, password).get_json()
    assert_retrieving_current_user_works(client, expected=user_response)

    # Getting a user by id returns UserSchema.
    user_expected = {
        k: v for k, v in user_response.items() if k in ["username", "email", "id"]
    }
    assert_retrieving_user_by_id_works(client, user_expected["id"], user_expected)


def test_user_get_all(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_users: dict[str, Any],
) -> None:
    """Test that all users can be retrieved.

    Given an authenticated user and registered users, this test validates the following
    sequence of actions:

    - Three users are registered, "user1", "user2", "user3".
    - The user is able to retrieve a list of all registered users.
    - The returned list of users matches the full list of registered users.
    """
    user_expected_list = [
        {"username": user["username"], "email": user["email"], "id": user["id"]}
        for user in list(registered_users.values())
    ]
    assert_retrieving_users_works(client, expected=user_expected_list)


def test_user_search_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_users: dict[str, Any],
) -> None:
    """Test that users can be found using a search query.

    Given an authenticated user and registered users, this test validates the following
    sequence of actions:

    - The user is able to retrieve a list of all users that match the query.
    - The returned list of users matches the expected matches from the query.
    """
    user_expected_list = [
        {"username": user["username"], "email": user["email"], "id": user["id"]}
        for user in list(registered_users.values())[:2]
    ]
    assert_retrieving_users_works(
        client, expected=user_expected_list, search="username:*user*"
    )
    assert_retrieving_users_works(
        client, expected=user_expected_list, search="username:'*user*'"
    )
    assert_retrieving_users_works(
        client, expected=user_expected_list, search='username:"user?"'
    )
    assert_retrieving_users_works(
        client, expected=user_expected_list, search="username:user?,email:user*"
    )
    assert_retrieving_users_works(client, expected=[], search=r"username:\*user*")
    assert_retrieving_users_works(client, expected=[], search="email:user?")


def test_cannot_register_existing_username(
    client: FlaskClient,
    db: SQLAlchemy,
    registered_users: dict[str, Any],
) -> None:
    """Test that registering a user with an existing username fails.

    Given an authenticated user and registered users, this test validates the following
    sequence of actions:

    - One user is registered, "user1".
    - A new user is registered with the same email as user1, a different username.
    - The request fails with an appropriate error message and response code.
    """
    existing_user = registered_users["user1"]
    assert_registering_existing_username_fails(
        client,
        existing_username=existing_user["username"],
        non_existing_email="unique" + existing_user["email"],
    )


def test_cannot_register_existing_email(
    client: FlaskClient,
    db: SQLAlchemy,
    registered_users: dict[str, Any],
) -> None:
    """Test that registering a user with an existing email fails.

    Given an authenticated user and registered users, this test validates the following
    sequence of actions:

    - One user is registered, "user1".
    - A new user is registered with the same email as user1, but _not_ the same email.
    - The request fails with an appropriate error message and response code.
    """
    existing_user = registered_users["user1"]
    assert_registering_existing_email_fails(
        client,
        non_existing_username="unique" + existing_user["username"],
        existing_email=existing_user["email"],
    )


def test_rename_current_user(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> None:
    """Test that renaming the current user works.

    Given an authenticated user, this test validates the following sequence of actions:

    - A user modifies its username and we record the response from the API.
    - Retrieving the current user works and responds with the response we recorded
    that reflects the updated username.
    """
    new_username = "new_name"
    user = modify_current_user(client, new_username, auth_account["email"]).get_json()
    assert_retrieving_current_user_works(client, expected=user)


def test_user_authorization_failure(
    client: FlaskClient,
    db: SQLAlchemy,
    registered_users: dict[str, Any],
) -> None:
    """Test that a user providing an incorrect password cannot log in.

    Given an authenticated user, this test validates the following sequence of actions:

    - The current attempts to login with an incorrect password.
    - The user cannot log in.
    """
    username = registered_users["user2"]["username"]
    password = registered_users["user2"]["password"] + "incorrect"
    assert_login_is_unauthorized(client, username=username, password=password)


def test_delete_current_user(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> None:
    """Test that deleting the current user works.

    Given an authenticated user, this test validates the following sequence of actions:

    - The current user deletes itself using its password.
    - The user can no longer log in.
    """
    username = auth_account["username"]
    password = auth_account["password"]
    delete_current_user(client, password)
    assert_user_does_not_exist(client, username=username, password=password)


def test_change_current_user_password(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
):
    """Test that changing the current user password works.

    Given an authenticated user, this test validates the following sequence of actions:

    - The current user changes its password, and it does not fail.
    - The current user is able to login with the new password.
    """
    username = auth_account["username"]
    old_password = auth_account["password"]
    new_password = "new_password"
    change_current_user_password(client, old_password, new_password)
    assert_login_works(client, username=username, password=new_password)


def test_change_user_password(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_users: dict[str, Any],
):
    """Test that changing a user password works.

    Given an authenticated user, this test validates the following sequence of actions:

    - Using its ID, a user's password is changed, and it does not fail.
    - This user is then able to login with the new password.
    """
    user_id = registered_users["user2"]["id"]
    username = registered_users["user2"]["username"]
    old_password = registered_users["user2"]["password"]
    new_password = "new_password"
    change_user_password(client, user_id, old_password, new_password)
    assert_login_works(client, username=username, password=new_password)


def test_new_password_cannot_be_existing(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
):
    """Test that changing a password and setting the new password as the
    old password fails.

    Given an authenticated user, this test validates the following sequence of actions:

    - The current user tries to set its new password which is the same as its existing
    password and this action fails.
    - The same password as the existing password for a user with an ID is attempted
    to be set and this action fails.
    """
    user_id = auth_account["id"]
    password = auth_account["password"]
    # test via /users/current
    assert_new_password_cannot_be_existing(client, password)
    # test via /users/{user_id}
    assert_new_password_cannot_be_existing(client, password, user_id)
